from __future__ import division

import itertools
import re
from collections import MutableSequence
from enum import Enum
from six import iteritems, string_types

from .list import shell_list
from .. import iterutils
from ..safe_str import jbos, safe_str, shell_literal

__all__ = ['split', 'join', 'listify', 'inner_quote', 'inner_quote_info',
           'wrap_quotes', 'quote', 'quote_info', 'escape_line', 'join_lines',
           'global_env']

# XXX: We need a way to escape cmd.exe-specific characters.
_bad_chars = re.compile(r'(\s|"|\\$)')
_replace = re.compile(r'(\\*)("|$)')
_ends_unescaped_quote = re.compile(r'(^|[^\\])(\\\\)*"$')

_Token = Enum('Token', ['char', 'quote', 'space'])
_State = Enum('State', ['between', 'char', 'word', 'quoted'])


def _tokenize(s):
    escapes = 0
    for c in s:
        if c == '\\':
            escapes += 1
        elif c == '"':
            for i in range(escapes // 2):
                yield (_Token.char, type(s)('\\'))
            yield (_Token.char, '"') if escapes % 2 else (_Token.quote, None)
            escapes = 0
        else:
            for i in range(escapes):
                yield (_Token.char, type(s)('\\'))
            yield ((_Token.space if c in ' \t' else _Token.char), c)
            escapes = 0


def split(s, type=list):
    if not isinstance(s, string_types):
        raise TypeError('expected a string')

    mutable = isinstance(type, MutableSequence)
    state = _State.between
    args = (type if mutable else list)()

    for tok, value in _tokenize(s):
        if state == _State.between:
            if tok == _Token.char:
                args.append(value)
                state = _State.word
            elif tok == _Token.quote:
                args.append('')
                state = _State.quoted
        elif state == _State.word:
            if tok == _Token.quote:
                state = _State.quoted
            elif tok == _Token.char:
                args[-1] += value
            else:  # tok == _Token.space
                state = _State.between
        else:  # state == _State.quoted
            if tok == _Token.quote:
                state = _State.word
            else:
                args[-1] += value

    return args if mutable else type(args)


def join(args):
    return ' '.join(quote(i) for i in args)


def listify(thing, type=list):
    if isinstance(thing, string_types):
        return split(thing, type)
    return iterutils.listify(thing, type=type)


def inner_quote_info(s, escape_percent=False):
    # In some contexts (mainly certain uses of the Windows shell), we want to
    # escape percent signs. This doesn't count as "escaping" for the purposes
    # of quoting the result though.
    if escape_percent:
        s = s.replace('%', '%%')

    if not _bad_chars.search(s):
        return s, False

    def repl(m):
        quote = '\\' + m.group(2) if len(m.group(2)) else ''
        return m.group(1) * 2 + quote
    return _replace.sub(repl, s), True


def inner_quote(s, escape_percent=False):
    return inner_quote_info(s, escape_percent)[0]


def wrap_quotes(s):
    # Thanks to `inner_quote_info` above, we can guarantee that any double-
    # quotes are escaped, so we don't need to worry about duplicates if they're
    # at the end.
    return '"' + s + '"'


def quote_info(s, escape_percent=False):
    s, quoted = inner_quote_info(s, escape_percent)
    return (wrap_quotes(s), True) if quoted else (s, False)


def quote(s, escape_percent=False):
    return quote_info(s, escape_percent)[0]


def escape_line(line, listify=False):
    if iterutils.isiterable(line):
        return iterutils.listify(line) if listify else line

    line = safe_str(line)
    if isinstance(line, string_types):
        line = shell_literal(line)
    elif isinstance(line, jbos):
        line = jbos(*(shell_literal(i) if isinstance(i, string_types) else i
                      for i in line.bits))
    return shell_list([line])


def join_lines(lines):
    result = []
    for i in iterutils.tween( (escape_line(j, listify=True) for j in lines),
                              shell_list([shell_literal('&&')]) ):
        result += i
    return result


def global_env(env, lines=[]):
    # Join the name and value so they get quoted together, if necessary.
    env_vars = (shell_list(['set', safe_str(name) + '=' + safe_str(value)])
                for name, value in iteritems(env))
    return join_lines(itertools.chain(env_vars, lines))



from ..backends.make import writer as make
from ..backends.ninja import writer as ninja
from ..build_inputs import build_input
from ..iterutils import listify
from ..path import Path


@build_input('regenerate')
class Regenerate(object):
    def __init__(self, build_inputs, env):
        self.outputs = []
        self.depfile = None


@make.post_rule
def make_regenerate_rule(build_inputs, buildfile, env):
    bfg9000 = env.tool('bfg9000')

    make.multitarget_rule(
        buildfile,
        targets=[Path('Makefile')] + build_inputs['regenerate'].outputs,
        deps=build_inputs.bootstrap_paths + listify(env.toolchain.path),
        recipe=[bfg9000(Path('.'))]
    )


@ninja.post_rule
def ninja_regenerate_rule(build_inputs, buildfile, env):
    bfg9000 = env.tool('bfg9000')

    buildfile.rule(
        name='regenerate',
        command=bfg9000(Path('.')),
        generator=True,
        depfile=build_inputs['regenerate'].depfile,
    )
    buildfile.build(
        output=[Path('build.ninja')] + build_inputs['regenerate'].outputs,
        rule='regenerate',
        implicit=build_inputs.bootstrap_paths + listify(env.toolchain.path)
    )



import os

from ... import iterutils
from ... import path
from ... import shell
from .syntax import *
from ...versioning import SpecifierSet, Version


def version(env=os.environ):
    try:
        ninja = shell.which(env.get('NINJA', ['ninja', 'ninja-build']), env)
        output = shell.execute(ninja + ['--version'], stdout=shell.Mode.pipe,
                               stderr=shell.Mode.devnull)
        return Version(output.strip())
    except (IOError, OSError, shell.CalledProcessError):
        pass
    return None


priority = 3
filepath = path.Path('build.ninja')

_rule_handlers = {}
_pre_rules = []
_post_rules = []


def rule_handler(*args):
    def decorator(fn):
        for i in args:
            _rule_handlers[i] = fn
        return fn
    return decorator


def pre_rule(fn):
    _pre_rules.append(fn)
    return fn


def post_rule(fn):
    _post_rules.append(fn)
    return fn


def write(env, build_inputs):
    buildfile = NinjaFile(build_inputs.bfgpath.string(env.base_dirs))
    buildfile.variable(path_vars[path.Root.srcdir], env.srcdir, Section.path)

    for i in _pre_rules:
        i(build_inputs, buildfile, env)
    for e in build_inputs.edges():
        _rule_handlers[type(e)](e, build_inputs, buildfile, env)
    for i in _post_rules:
        i(build_inputs, buildfile, env)

    with open(filepath.string(env.base_dirs), 'w') as out:
        buildfile.write(out)


def flags_vars(name, value, buildfile):
    gflags = buildfile.variable('global_' + name, value, Section.flags, True)
    flags = buildfile.variable(name, gflags, Section.other, True)
    return gflags, flags


def command_build(buildfile, env, output, inputs=None, implicit=None,
                  order_only=None, command=[], console=False, phony=False,
                  description=None):
    if phony:
        extra_implicit = ['PHONY']
        if not buildfile.has_build('PHONY'):
            buildfile.build(output='PHONY', rule='phony')
    else:
        extra_implicit = []

    if ( console and env.backend_version and env.backend_version in
         SpecifierSet('>=1.5') ):
        rule_name = 'console_command'
        rule_kwargs = {'pool': 'console'}
    else:
        rule_name = 'command'
        rule_kwargs = {}

    if not buildfile.has_rule(rule_name):
        buildfile.rule(name=rule_name, command=shell.shell_list([var('cmd')]),
                       **rule_kwargs)

    variables = {'cmd': command}
    if description:
        variables['description'] = description
    buildfile.build(
        output=output,
        rule=rule_name,
        inputs=inputs,
        implicit=iterutils.listify(implicit) + extra_implicit,
        order_only=order_only,
        variables=variables
    )



import subprocess

from .app_version import version
from .arguments import parser as argparse
from .iterutils import tween


def make_depfile(subcmd, output, depfile):
    deps = subprocess.check_output(subcmd + ['--list'],
                                   universal_newlines=True)
    deps = deps.strip().split('\n')

    with open(depfile, 'w') as f:
        f.write(output + ': ')
        for i in tween(deps, ' \\\n'):
            f.write(i)
        f.write('\n')


def run_rcc(subcmd, output):
    return subprocess.call(subcmd + ['-o', output])


def main():
    parser = argparse.ArgumentParser(
        prog='bfg9000-rccdep',
        usage='%(prog)s SUBCMD -o FILE -d FILE',
        description="Write a depfile (in Makefile syntax) for Qt's rcc " +
                    "tool, and then run rcc."
    )
    parser.add_argument('--version', action='version',
                        version='%(prog)s ' + version)
    parser.add_argument('-o', '--output', required=True, metavar='FILE',
                        help='the output file for rcc (required)')
    parser.add_argument('-d', '--depfile', required=True, metavar='FILE',
                        help='the depfile to generate (required)')
    args, subcommand = parser.parse_known_args()

    subparser = argparse.ArgumentParser(prog=parser.prog, usage=parser.usage)
    subparser.add_argument('command', nargs='+')
    subargs = subparser.parse_known_args(subcommand)[0]

    try:
        make_depfile(subargs.command, args.output, args.depfile)
        return run_rcc(subcommand, args.output)
    except Exception as e:
        parser.error(e)



from .core import _get_platform_info, _platform_info, Platform


class HostPlatform(Platform):
    pass


def platform_info(*args, **kwargs):
    return _platform_info('host', *args, **kwargs)


def from_json(value):
    return _get_platform_info('host', value['genus'], value['species'],
                              value['arch'])



from six import with_metaclass as _with_metaclass

try:
    from enum import EnumMeta as _EnumMeta, Flag as _Flag
except ImportError:
    from enum import EnumMeta as _EnumMeta, IntEnum as _Flag

from .options import option_list as _option_list
from .platforms.framework import Framework  # noqa


class _PackageKindMeta(_EnumMeta):
    def __getitem__(cls, name):
        try:
            return _EnumMeta.__getitem__(cls, name)
        except KeyError:
            pass
        raise ValueError('kind must be one of: {}'.format(', '.join(
            "'{}'".format(i.name) for i in cls
        )))


class PackageKind(_with_metaclass(_PackageKindMeta, _Flag)):
    static = 1
    shared = 2
    any = static | shared


class Package(object):
    is_package = True

    def __init__(self, name, format):
        self.name = name
        self.format = format

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, rhs):
        return (type(self) == type(rhs) and self.name == rhs.name and
                self.format == rhs.format)

    def __ne__(self, rhs):
        return not (self == rhs)

    def __repr__(self):
        return '<{type} {name}>'.format(
            type=type(self).__name__, name=repr(self.name)
        )


class CommonPackage(Package):
    def __init__(self, name, format, compile_options=None, link_options=None):
        Package.__init__(self, name, format)
        self._compile_options = compile_options or _option_list()
        self._link_options = link_options or _option_list()

    def compile_options(self, compiler):
        return self._compile_options

    def link_options(self, linker):
        return self._link_options



class shell_list(list):
    """A special subclass of list used to mark that this command line uses
    special shell characters."""

    def __repr__(self):
        return '<shell_list({})>'.format(list.__repr__(self))

    def __add__(self, rhs):
        return shell_list(list.__add__(self, rhs))

    def __radd__(self, lhs):
        return shell_list(list.__add__(lhs, self))

    def __getitem__(self, key):
        return shell_list(list.__getitem__(self, key))

    def __getslice__(self, i, j):
        return shell_list(list.__getslice__(self, i, j))

    def __eq__(self, rhs):
        return isinstance(rhs, shell_list) and list.__eq__(self, rhs)

    def __ne__(self, rhs):
        return not self == rhs



from . import builtin
from ..backends.make import writer as make
from ..backends.ninja import writer as ninja
from ..build_inputs import Edge
from ..file_types import Phony


class Alias(Edge):
    def __init__(self, build, name, deps=None):
        Edge.__init__(self, build, Phony(name), extra_deps=deps)


@builtin.function('build_inputs')
def alias(build, *args, **kwargs):
    return Alias(build, *args, **kwargs).public_output


@make.rule_handler(Alias)
def make_alias(rule, build_inputs, buildfile, env):
    buildfile.rule(
        target=rule.output,
        deps=rule.extra_deps,
        phony=True
    )


@ninja.rule_handler(Alias)
def ninja_alias(rule, build_inputs, buildfile, env):
    buildfile.build(
        output=rule.output,
        rule='phony',
        inputs=rule.extra_deps
    )


try:
    from ..backends.msbuild import writer as msbuild

    @msbuild.rule_handler(Alias)
    def msbuild_alias(rule, build_inputs, solution, env):
        output = rule.output[0]
        project = msbuild.NoopProject(
            env, name=output.path,
            dependencies=solution.dependencies(rule.extra_deps),
        )
        solution[output] = project
except ImportError:  # pragma: no cover
    pass



import os
import tarfile

from . import *
pjoin = os.path.join


class TestSubdirs(IntegrationTest):
    def __init__(self, *args, **kwargs):
        IntegrationTest.__init__(self, pjoin(examples_dir, '05_subdirs'),
                                 install=True, *args, **kwargs)

    def setUp(self):
        IntegrationTest.setUp(self)
        cleandir(self.installdir)

    def test_build(self):
        self.build()
        self.assertOutput([executable('sub/program')], 'hello, library!\n')

    def _check_installed(self):
        extra = []
        if env.target_platform.has_import_library:
            extra = [pjoin(self.libdir, import_library('sub/library').path)]

        self.assertDirectory(self.installdir, [
            pjoin(self.includedir, 'library.hpp'),
            pjoin(self.includedir, 'detail', 'export.hpp'),
            pjoin(self.bindir, executable('sub/program').path),
            pjoin(self.libdir, shared_library('sub/library').path),
        ] + extra)

    @skip_if_backend('msbuild')
    def test_dist(self):
        dist = output_file('05_subdirs.tar.gz')
        self.build('dist')
        self.assertExists(dist)
        with tarfile.open(self.target_path(dist)) as t:
            self.assertEqual(set(t.getnames()), {
                '05_subdirs/build.bfg',
                '05_subdirs/include',
                '05_subdirs/include/library.hpp',
                '05_subdirs/include/detail/export.hpp',
                '05_subdirs/src/library.cpp',
                '05_subdirs/src/program.cpp',
            })

    @only_if_backend('make', hide=True)
    def test_dir_sentinels(self):
        self.build()
        self.assertTrue(os.path.isfile('sub/.dir'))

    @skip_if_backend('msbuild')
    def test_install(self):
        self.build('install')
        self._check_installed()

        os.chdir(self.srcdir)
        cleandir(self.builddir)
        self.assertOutput([pjoin(self.bindir, executable('sub/program').path)],
                          'hello, library!\n')

    @skip_if_backend('msbuild')
    def test_install_existing_paths(self):
        makedirs(self.includedir, exist_ok=True)
        makedirs(self.bindir, exist_ok=True)
        makedirs(self.libdir, exist_ok=True)
        self.build('install')
        self._check_installed()

        os.chdir(self.srcdir)
        cleandir(self.builddir)
        self.assertOutput([pjoin(self.bindir, executable('sub/program').path)],
                          'hello, library!\n')

    @skip_if_backend('msbuild')
    def test_uninstall(self):
        self.build('install')
        self._check_installed()

        self.build('uninstall')
        self.assertDirectory(self.installdir, [])



from itertools import chain

from . import tool
from .common import SimpleCommand
from ..iterutils import iterate


@tool('rm')
class Rm(SimpleCommand):
    def __init__(self, env):
        default = ['rm -f']
        if env.host_platform.family == 'windows':
            default.append('cmd /c del')
        SimpleCommand.__init__(self, env, name='rm', env_var='RM',
                               default=default)

    def _call(self, cmd, files):
        return list(chain(cmd, iterate(files)))



from . import builtin
from .. import options
from ..objutils import memoize


class Options(object):
    def __init__(self, options):
        self._options = options

    def __getattr__(self, key):
        return self._options[key]


@builtin.getter()
@memoize
def opts():
    return Options(options.Option.registry)



from six import iteritems

from . import tool
from .common import SimpleCommand
from ..platforms.host import platform_info
from ..safe_str import jbos, safe_str, shell_literal
from ..shell import escape_line


if platform_info().family == 'windows':
    @tool('setenv')
    class SetEnv(SimpleCommand):
        def __init__(self, env):
            SimpleCommand.__init__(self, env, name='setenv', env_var='SETENV',
                                   default=env.bfgdir.append('pysetenv'))

        def _call(self, cmd, env, line):
            if env:
                eq = shell_literal('=')
                env_vars = cmd + [jbos(safe_str(name), eq, safe_str(value))
                                  for name, value in iteritems(env)] + ['--']
            else:
                env_vars = []
            return env_vars + escape_line(line, listify=True)



import functools
import inspect
import sys
from itertools import chain
from six import iteritems, iterkeys, itervalues, string_types

from ..iterutils import iterate, listify
from ..platforms.basepath import BasePath

string_or_path_types = string_types + (BasePath,)


class Builtins(object):
    def __init__(self):
        self._default = {}
        self._post = {}

    def add(self, kind, name, value):
        assert kind in ('default', 'post')
        getattr(self, '_' + kind)[name] = value

    def bind(self, **kwargs):
        builtins = {}
        for k, v in iteritems(self._default):
            builtins[k] = v.bind(builtins=builtins, **kwargs)

        builtins['__bfg9000__'] = builtins
        return builtins

    def run_post(self, builtins, **kwargs):
        for v in itervalues(self._post):
            v(builtins=builtins, **kwargs)


build = Builtins()
options = Builtins()
toolchain = Builtins()
_allbuiltins = {
    'build': build,
    'options': options,
    'toolchain': toolchain,
}


def _add_builtin(context, kind, name, value):
    if context == '*':
        context = iterkeys(_allbuiltins)
    for i in iterate(context):
        _allbuiltins[i].add(kind, name, value)


class _Binder(object):
    def __init__(self, fn):
        self._fn = fn

    def bind(self, **kwargs):
        return self._fn


class _PartialFunctionBinder(_Binder):
    def __init__(self, fn, *args):
        _Binder.__init__(self, fn)
        self._args = args

    def bind(self, **kwargs):
        if not self._args:
            return _Binder.bind(self, **kwargs)
        pre_args = tuple(kwargs[i] for i in self._args)

        @functools.wraps(self._fn)
        def wrapper(*args, **kwargs):
            return self._fn(*(pre_args + args), **kwargs)

        if sys.version_info >= (3, 3):
            sig = inspect.signature(wrapper)
            params = list(sig.parameters.values())[len(kwargs):]
            wrapper.__signature__ = inspect.Signature(params)
        return wrapper


class _GetterBinder(_Binder):
    def __init__(self, fn, *args):
        _Binder.__init__(self, fn)
        self._args = args

    def bind(self, **kwargs):
        return self._fn(*[kwargs[i] for i in self._args])


class _PostWrapper(object):
    def __init__(self, fn, *args):
        self._fn = fn
        self._args = args

    def __call__(self, **kwargs):
        args = tuple(kwargs[i] for i in self._args)
        return self._fn(*args)


class _Decorator(object):
    def __init__(self, kind, binder):
        self.__kind = kind
        self.__binder = binder

    def __call__(self, *args, **kwargs):
        context = kwargs.pop('context', 'build')
        name = kwargs.pop('name', None)

        def decorator(fn):
            _add_builtin(context, self.__kind, name or fn.__name__,
                         self.__binder(fn, *args))
            fn._builtin_bound = len(args)
            return fn
        return decorator


function = _Decorator('default', _PartialFunctionBinder)
getter = _Decorator('default', _GetterBinder)
post = _Decorator('post', _PostWrapper)


def _get_argspec(fn):
    if sys.version_info >= (3, 3):
        return list(inspect.signature(fn).parameters.keys())
    return inspect.getargspec(fn).args


def _get_value(argspec, index, args, kwargs):
    # Get the value of the nth argument to this function, whether it's
    # passed positionally or as a keyword argument. Note that `index` should be
    # at least as large as the number of builtins bound to the function.
    if len(args) > index:
        return args[index]
    name = argspec[index]
    if name in kwargs:
        return kwargs[name]
    raise IndexError('unable to find user-provided argument')


def check_types(thing, expected_types, extra_types=[]):
    if not isinstance(thing, expected_types):
        types = chain(extra_types, expected_types)
        raise TypeError('expected {}; but got {}'.format(
            ', '.join(i.__name__ for i in types),
            __builtins__['type'](thing).__name__
        ))


def type(out_type, in_type=string_or_path_types, extra_in_type=(),
         short_circuit=True, first_optional=False):
    in_type = listify(in_type, type=tuple) + listify(extra_in_type, type=tuple)
    if first_optional:
        in_type = in_type + (__builtins__['type'](None),)

    def decorator(fn):
        spec = _get_argspec(fn)
        all_types = (out_type,) + in_type

        @functools.wraps(fn)
        def wrapper(*args, **kwargs):
            bound = getattr(wrapper, '_builtin_bound', 0)
            if first_optional and len(args) < 2 + bound:
                args = args[:bound] + (None, ) + args[bound:]

            # Try to get the first argument to this function. If it's the
            # output type, just return it immediately; otherwise, check if it's
            # a valid input type and then call the function.
            try:
                thing = _get_value(spec, bound, args, kwargs)
                if short_circuit:
                    if isinstance(thing, wrapper.type):
                        return thing
                    check_types(thing, wrapper.in_type, [wrapper.type])
                else:
                    check_types(thing, all_types)
            except IndexError:
                pass
            return fn(*args, **kwargs)

        wrapper.type = out_type
        wrapper.in_type = in_type
        return wrapper
    return decorator



from . import tool
from .common import SimpleCommand


class LinkCommand(SimpleCommand):
    def __init__(self, *args, **kwargs):
        SimpleCommand.__init__(self, *args, **kwargs)
        lower_cmd = [i.lower() for i in self.command]
        self.flavor = 'mklink' if 'mklink' in lower_cmd else 'ln'

    def _call(self, cmd, input, output):
        if self.flavor == 'mklink':
            return cmd + [output, input]
        return cmd + [input, output]


@tool('symlink')
class Symlink(LinkCommand):
    def __init__(self, env):
        default = ('cmd /c mklink' if env.host_platform.family == 'windows'
                   else 'ln -sf')
        LinkCommand.__init__(self, env, name='symlink', env_var='SYMLINK',
                             default=default)

    def transform_input(self, input, output):
        try:
            return input.path.relpath(output.path.parent())
        except ValueError:
            return input.path


@tool('hardlink')
class Hardlink(LinkCommand):
    def __init__(self, env):
        default = ('cmd /c mklink /H' if env.host_platform.family == 'windows'
                   else 'ln -f')
        LinkCommand.__init__(self, env, name='hardlink', env_var='HARDLINK',
                             default=default)


@tool('copy')
class Copy(SimpleCommand):
    def __init__(self, env):
        default = ('cmd /c copy' if env.host_platform.family == 'windows'
                   else 'cp -f')
        SimpleCommand.__init__(self, env, name='cp', env_var='CP',
                               default=default)

    def _call(self, cmd, input, output):
        return cmd + [input, output]



import os
import shutil
import subprocess
import time
import unittest
from collections import namedtuple
from six import iteritems

from .. import *

from bfg9000.backends import list_backends
from bfg9000.iterutils import listify
from bfg9000.path import InstallRoot, makedirs, Path, Root

this_dir = os.path.abspath(os.path.dirname(__file__))
examples_dir = os.path.join(this_dir, '..', '..', 'examples')
test_data_dir = os.path.join(this_dir, '..', 'data')
test_stage_dir = os.path.join(this_dir, '..', 'stage')

env = make_env()

Target = namedtuple('Target', ['name', 'path'])

if os.getenv('BACKENDS', '').strip():
    backends = os.getenv('BACKENDS').split(' ')
else:
    backends = [k for k, v in iteritems(list_backends()) if v.priority > 0]
    # Only test with MSBuild by default on Windows.
    if env.host_platform.family != 'windows' and 'msbuild' in backends:
        backends.remove('msbuild')

extra_tests = os.getenv('BFG_EXTRA_TESTS', '').split(' ')
skipped_tests = os.getenv('BFG_SKIPPED_TESTS', '').split(' ')

_unset = object()


def cleandir(path, recreate=True):
    if os.path.exists(path):
        # Windows seems to keep an executable file open a little bit after the
        # process returns from wait(), so try a few times, sleeping a bit in
        # between.
        for t in [0.1, 0.25, 0.5, 1.0, 2.0, None]:
            try:
                shutil.rmtree(path)
                break
            except OSError:
                if t is None:
                    raise RuntimeError('unable to remove {}'.format(path))
                time.sleep(t)
    if recreate:
        makedirs(path)


def skip_if_backend(backend, hide=False):
    return skip_pred(lambda x: x.backend == backend,
                     'not supported for backend "{}"'.format(backend), hide)


def only_if_backend(backend, hide=False):
    return skip_pred(lambda x: x.backend != backend,
                     'only supported for backend "{}"'.format(backend), hide)


class SubprocessError(unittest.TestCase.failureException):
    def __init__(self, returncode, env, message):
        envstr = ''.join('  {} = {}\n'.format(k, v)
                         for k, v in iteritems(env or {}))
        unittest.TestCase.failureException.__init__(
            self, 'returned {returncode}\n{env}{line}\n{msg}\n{line}'.format(
                returncode=returncode, env=envstr, line='-' * 60, msg=message
            )
        )


class SubprocessTestCase(TestCase):
    def target_name(self, target):
        if self.backend == 'msbuild':
            if isinstance(target, Target):
                target = target.name
            return '/target:' + target
        else:
            if isinstance(target, Target):
                target = target.path
            return target

    def target_path(self, target):
        if isinstance(target, Target):
            prefix = '.'
            if self.backend == 'msbuild':
                prefix = 'Debug'
                if os.getenv('PLATFORM') == 'X64':
                    prefix = os.path.join('X64', prefix)
            return os.path.join(prefix, target.path)
        return target

    def assertPopen(self, command, env=None, env_update=True, input=None,
                    returncode=0):
        final_env = env
        if env is not None and env_update:
            final_env = dict(os.environ)
            final_env.update(env)

        command = [self.target_path(i) for i in command]
        proc = subprocess.Popen(
            command, stdin=subprocess.PIPE, stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT, env=final_env, universal_newlines=True
        )
        output = proc.communicate(input)[0]
        if not (returncode == 'any' or
                (returncode == 'fail' and proc.returncode != 0) or
                proc.returncode in listify(returncode)):
            raise SubprocessError(proc.returncode, env, output)
        return output

    def assertOutput(self, command, output, *args, **kwargs):
        self.assertEqual(self.assertPopen(command, *args, **kwargs), output)

    def assertExists(self, path):
        realpath = self.target_path(path)
        if not os.path.exists(realpath):
            raise unittest.TestCase.failureException(
                "'{}' does not exist".format(realpath)
            )

    def assertNotExists(self, path):
        realpath = self.target_path(path)
        if os.path.exists(realpath):
            raise unittest.TestCase.failureException(
                "'{}' exists".format(os.path.normpath(realpath))
            )

    def assertDirectory(self, path, contents):
        path = os.path.normpath(path)
        actual = set(os.path.normpath(os.path.join(path, base, f))
                     for base, dirs, files in os.walk(path) for f in files)
        expected = set(os.path.normpath(os.path.join(path, i))
                       for i in contents)
        if actual != expected:
            missing = [os.path.relpath(i, path) for i in (expected - actual)]
            extra = [os.path.relpath(i, path) for i in (actual - expected)]
            raise unittest.TestCase.failureException(
                'missing: {}, extra: {}'.format(missing, extra)
            )


class BasicIntegrationTest(SubprocessTestCase):
    def __init__(self, srcdir, *args, **kwargs):
        install = kwargs.pop('install', False)
        self._configure = kwargs.pop('configure', True)

        stage_src = kwargs.pop('stage_src', False)
        self.backend = kwargs.pop('backend', None)
        self.env = kwargs.pop('env', None)

        self.extra_args = kwargs.pop('extra_args', [])

        SubprocessTestCase.__init__(self, *args, **kwargs)
        if self.backend is None:
            return

        srcname = os.path.basename(srcdir)
        self.srcdir = os.path.join(test_data_dir, srcdir)
        if stage_src:
            srcname = os.path.basename(srcdir)
            self.orig_srcdir = self.srcdir
            self.srcdir = os.path.join(test_stage_dir, srcname)
        else:
            self.orig_srcdir = None

        if install:
            self.installdir = os.path.join(test_stage_dir,
                                           srcname + '-install')

            install_dirs = env.target_platform.install_dirs.copy()
            install_dirs[InstallRoot.prefix] = Path(
                self.installdir, Root.absolute
            )
            for i in InstallRoot:
                setattr(self, i.name, install_dirs[i].string(install_dirs))
        else:
            self.installdir = None

    @staticmethod
    def _make_builddir(srcdir):
        srcname = os.path.basename(srcdir)
        return os.path.join(test_stage_dir, srcname + '-build')

    @property
    def builddir(self):
        return self._make_builddir(self.srcdir)

    def parameterize(self):
        return ([] if self._hideTest() else
                [self.__class__(backend='', *self._args, **self._kwargs)])

    def shortDescription(self):
        return self.backend

    def setUp(self):
        if self._configure:
            self.configure()

    def configure(self, srcdir=None, builddir=None, installdir=_unset,
                  orig_srcdir=_unset, extra_args=_unset, env=_unset,
                  backend=_unset, returncode=0):
        if srcdir:
            srcdir = os.path.join(test_data_dir, srcdir)
        else:
            srcdir = self.srcdir
        builddir = builddir or self._make_builddir(srcdir)

        if installdir is _unset:
            installdir = self.installdir
        if orig_srcdir is _unset:
            orig_srcdir = self.orig_srcdir
        if extra_args is _unset:
            extra_args = self.extra_args
        if env is _unset:
            env = self.env
        if backend is _unset:
            backend = self.backend

        if orig_srcdir:
            cleandir(srcdir, recreate=False)
            shutil.copytree(orig_srcdir, srcdir)
        os.chdir(srcdir)
        cleandir(builddir)

        if installdir:
            cleandir(installdir)
            install_args = ['--prefix', installdir]
        else:
            install_args = []

        self.assertPopen(
            ['bfg9000', '--debug', 'configure', builddir,
             '--backend', backend] + install_args + extra_args,
            env=env, env_update=True, returncode=returncode
        )
        os.chdir(builddir)

    def build(self, target=None, extra_args=[]):
        args = [os.getenv(self.backend.upper(), self.backend)] + extra_args
        if target:
            args.append(self.target_name(target))
        return self.assertPopen(args)

    def clean(self):
        if self.backend == 'ninja':
            return self.build(extra_args=['-t', 'clean'])
        elif self.backend == 'make':
            return self.build('clean')
        else:  # self.backend == 'msbuild'
            return self.build(extra_args=['/t:Clean'])

    def wait(self, t=1):
        time.sleep(t)


class IntegrationTest(BasicIntegrationTest):
    def parameterize(self):
        return parameterize_tests(self, backend=backends)


def output_file(name):
    return Target(name, os.path.normpath(os.path.join('.', name)))


def executable(name):
    return Target(name, os.path.normpath(os.path.join(
        '.', name + env.target_platform.executable_ext
    )))


if env.builder('c++').flavor == 'msvc':
    _shared_library_prefix = ''
    _static_library_prefix = ''
elif env.target_platform.family == 'windows':
    _shared_library_prefix = ''
    _static_library_prefix = 'lib'
else:
    _shared_library_prefix = 'lib'
    _static_library_prefix = 'lib'


def shared_library(name, version=None):
    head, tail = os.path.split(name)
    ext = env.target_platform.shared_library_ext
    if version:
        if not env.target_platform.has_versioned_library:
            raise ValueError('no versioned libraries on this platform')
        if env.target_platform.genus == 'darwin':
            tail += '.' + version
        else:
            ext += '.' + version
    return Target(name, os.path.normpath(os.path.join(
        '.', head, _shared_library_prefix + tail + ext
    )))


def static_library(name):
    suffix = '.lib' if env.builder('c++').flavor == 'msvc' else '.a'
    head, tail = os.path.split(name)
    return Target(name, os.path.normpath(os.path.join(
        '.', head, _static_library_prefix + tail + suffix
    )))


def import_library(name):
    if not env.target_platform.has_import_library:
        raise ValueError('no import libraries on this platform')
    if env.builder('c++').flavor == 'msvc':
        return static_library(name)
    else:
        head, tail = os.path.split(shared_library(name).path)
        path = os.path.join(head, _static_library_prefix + tail + '.a')
        return Target(name, path)



import os
import re
import warnings
from six import iteritems

from .. import options as opts
from .. import path, shell
from ..iterutils import first, iterate, listify, slice_dict

_modes = {
    'shared_library': 'EXPORTS',
    'static_library': 'STATIC',
}

_macro_ex = re.compile(r'([^A-Za-z0-9_]|^(?=[0-9]))')
_macro_ex2 = re.compile(r'^_')


def library_macro(name, mode):
    # Replace all non-alphanumeric characters in the name with underscores and
    # prepend an underscore if the name starts with a digit. Then add an extra
    # 'LIB' to the beginning if the transformed name starts with an underscore.
    return '{name}_{suffix}'.format(
        name=_macro_ex2.sub('LIB_', _macro_ex.sub('_', name.upper())),
        suffix=_modes[mode]
    )


def not_buildroot(thing):
    if isinstance(thing, path.BasePath):
        return thing != type(thing)('.')
    return thing is not None


class Builder(object):
    def __init__(self, lang, brand, version):
        self.lang = lang
        self.brand = brand
        self.version = version

    def __repr__(self):
        return '<{}({!r})>'.format(type(self).__name__, self.brand)


class Command(object):
    def __init__(self, env, rule_name, command_var, command):
        self.env = env
        self.rule_name = rule_name
        self.command_var = command_var
        self.command = command

    @staticmethod
    def convert_args(args, conv):
        args = listify(args, scalar_ok=False)
        if not any(isinstance(i, Command) for i in args):
            return args

        result = type(args)()
        for i in args:
            if isinstance(i, Command):
                result.extend(listify(conv(i)))
            else:
                result.append(i)
        return result

    def __call__(self, *args, **kwargs):
        cmd = listify(kwargs.pop('cmd', self))
        return self._call(cmd, *args, **kwargs)

    def run(self, *args, **kwargs):
        run_kwargs = slice_dict(kwargs, ('env', 'env_update'))

        return self.env.execute(
            self(*args, **kwargs), stdout=shell.Mode.pipe,
            stderr=shell.Mode.devnull, **run_kwargs
        )

    def __repr__(self):
        return '<{}({})>'.format(
            type(self).__name__, ', '.join(repr(i) for i in self.command)
        )


class SimpleCommand(Command):
    def __init__(self, env, name, env_var, default, kind='executable'):
        cmd = check_which(env.getvar(env_var, default), env.variables,
                          kind=kind)
        Command.__init__(self, env, name, name, cmd)


class BuildCommand(Command):
    def __init__(self, builder, env, rule_name, command_var, command,
                 **kwargs):
        Command.__init__(self, env, rule_name, command_var, command)
        self.builder = builder

        # Fill in the names and values of the various flags needed for this
        # command, e.g. `flags` ('cflags', 'ldflags'), `libs` ('ldlibs'), etc.
        for k, v in iteritems(kwargs):
            setattr(self, '{}_var'.format(k), v[0])
            setattr(self, 'global_{}'.format(k), v[1])

    @property
    def lang(self):
        return self.builder.lang

    @property
    def family(self):
        return self.builder.family

    @property
    def brand(self):
        return self.builder.brand

    @property
    def version(self):
        return self.builder.version

    @property
    def flavor(self):
        return self.builder.flavor

    @property
    def num_outputs(self):
        return 'all'

    def pre_build(self, build, name, context):
        return opts.option_list()

    def post_build(self, build, options, output, context):
        return None

    def post_install(self, options, output, context):
        return None


def check_which(names, *args, **kwargs):
    names = listify(names)
    try:
        return shell.which(names, *args, **kwargs)
    except IOError as e:
        warnings.warn(str(e))
        # Assume the first name is the best choice.
        return shell.listify(names[0])


def choose_builder(env, langinfo, default_candidates, builders):
    candidates = listify(env.getvar(langinfo.var('compiler'),
                                    default_candidates))
    try:
        cmd = shell.which(candidates, env.variables,
                          kind='{} compiler'.format(langinfo.name))
    except IOError as e:
        warnings.warn(str(e))
        cmd = shell.listify(candidates[0])
        builder_type = first(builders)
        output = ''
    else:
        for builder_type in builders:
            try:
                output = builder_type.check_command(env, cmd)
                break
            except Exception:
                pass
        else:
            tried = ', '.join(repr(i) for i in iterate(candidates))
            raise IOError('no working {} compiler found; tried {}'
                          .format(langinfo.name, tried))

    return builder_type(env, langinfo, cmd, output)


def darwin_install_name(library):
    return os.path.join('@rpath', library.path.suffix)



import itertools
import re
from shlex import shlex
from six import iteritems, string_types

from .. import iterutils
from .list import shell_list
from ..platforms.host import platform_info
from ..safe_str import jbos, safe_str, shell_literal

__all__ = ['split', 'join', 'listify', 'inner_quote', 'inner_quote_info',
           'wrap_quotes', 'quote', 'quote_info', 'escape_line', 'join_lines',
           'local_env', 'global_env']

_bad_chars = re.compile(r'[^\w@%+=:,./-]')
_ends_unescaped_quote = re.compile(r"(^|[^\\])(\\\\)*'$")


def split(s, type=list, escapes=False):
    if not isinstance(s, string_types):
        raise TypeError('expected a string')
    lexer = shlex(s, posix=True)
    lexer.commenters = ''
    if not escapes:
        lexer.escape = ''
    lexer.whitespace_split = True
    return type(lexer)


def join(args):
    return ' '.join(quote(i) for i in args)


def listify(thing, type=list):
    if isinstance(thing, string_types):
        return split(thing, type)
    return iterutils.listify(thing, type=type)


def inner_quote_info(s):
    if _bad_chars.search(s):
        return s.replace("'", r"'\''"), True
    return s, False


def inner_quote(s):
    return inner_quote_info(s)[0]


def wrap_quotes(s):
    def q(has_quote):
        return '' if has_quote else "'"

    # Any string less than 3 characters long can't have escaped quotes.
    if len(s) < 3:
        return "'" + s + "'"

    start = 1 if s[0] == "'" else None
    # Thanks to `inner_quote_info` above, we can guarantee that any single-
    # quotes are unescaped, so we can deduplicate them if they're at the end.
    end = -1 if s[-1] == "'" else None
    return q(start) + s[start:end] + q(end)


def quote_info(s):
    s, quoted = inner_quote_info(s)
    return (wrap_quotes(s), True) if quoted else (s, False)


def quote(s):
    return quote_info(s)[0]


def _escape_word(word):
    # Since we can sometimes use an sh-style shell even on Windows (e.g.
    # with the Make backend), we want to escape backslashes when writing an
    # already "escaped" command line. Otherwise, Windows users would be
    # pretty surprised to find that all the paths they specified like
    # C:\foo\bar are broken!
    if platform_info().family == 'windows':
        word = word.replace('\\', '\\\\')
    return shell_literal(word)


def escape_line(line, listify=False):
    if iterutils.isiterable(line):
        return iterutils.listify(line) if listify else line

    line = safe_str(line)
    if isinstance(line, string_types):
        line = _escape_word(line)
    elif isinstance(line, jbos):
        line = jbos(*(_escape_word(i) if isinstance(i, string_types) else i
                      for i in line.bits))
    return shell_list([line])


def join_lines(lines):
    result = []
    for i in iterutils.tween( (escape_line(j, listify=True) for j in lines),
                              shell_list([shell_literal('&&')]) ):
        result += i
    return result


def local_env(env, line):
    if env:
        eq = shell_literal('=')
        env_vars = shell_list(jbos(safe_str(name), eq, safe_str(value))
                              for name, value in iteritems(env))
    else:
        env_vars = []
    return env_vars + escape_line(line, listify=True)


def global_env(env, lines=None):
    eq = shell_literal('=')
    env_vars = (shell_list([
        'export', jbos(safe_str(name), eq, safe_str(value))
    ]) for name, value in iteritems(env))
    return join_lines(itertools.chain(env_vars, lines or []))



from bfg9000 import safe_str

from . import *

jbos = safe_str.jbos
literal = safe_str.literal
shell_literal = safe_str.shell_literal


class MyString(object):
    def _safe_str(self):
        return 'foo'


class MyLiteral(object):
    def _safe_str(self):
        return literal('foo')


class MySafeStr(safe_str.safe_string):
    def __init__(self, i):
        self.i = i

    def __eq__(self, rhs):
        return type(self) == type(rhs) and self.i == rhs.i


class TestSafeStr(TestCase):
    def test_string(self):
        self.assertEqual(safe_str.safe_str('foo'), 'foo')

    def test_literals(self):
        self.assertEqual(safe_str.safe_str(literal('foo')), literal('foo'))
        self.assertEqual(safe_str.safe_str(shell_literal('foo')),
                         shell_literal('foo'))

    def test_jbos(self):
        self.assertEqual(safe_str.safe_str(jbos('foo')), jbos('foo'))

    def test_objects(self):
        self.assertEqual(safe_str.safe_str(MyString()), 'foo')
        self.assertEqual(safe_str.safe_str(MyLiteral()), literal('foo'))
        self.assertEqual(safe_str.safe_str(MySafeStr(1)), MySafeStr(1))

    def test_invalid(self):
        self.assertRaises(NotImplementedError, safe_str.safe_str, 123)


class TestLiteral(TestCase):
    def test_construct_from_string(self):
        self.assertEqual(literal('foo').string, 'foo')
        self.assertEqual(shell_literal('foo').string, 'foo')

    def test_construct_invalid(self):
        self.assertRaises(TypeError, literal, 123)
        self.assertRaises(TypeError, shell_literal, 123)

    def test_convert_to_str(self):
        with self.assertRaises(NotImplementedError):
            str(safe_str.literal('foo'))
        with self.assertRaises(NotImplementedError):
            str(safe_str.shell_literal('foo'))

    def test_equality(self):
        self.assertTrue(literal('foo') == literal('foo'))
        self.assertTrue(shell_literal('foo') == shell_literal('foo'))
        self.assertFalse(literal('foo') != literal('foo'))
        self.assertFalse(shell_literal('foo') != shell_literal('foo'))

        self.assertFalse(literal('foo') == literal('bar'))
        self.assertFalse(shell_literal('foo') == shell_literal('bar'))
        self.assertTrue(literal('foo') != literal('bar'))
        self.assertTrue(shell_literal('foo') != shell_literal('bar'))

        self.assertFalse(literal('foo') == shell_literal('foo'))
        self.assertFalse(shell_literal('foo') == literal('foo'))
        self.assertTrue(literal('foo') != shell_literal('foo'))
        self.assertTrue(shell_literal('foo') != literal('foo'))

    def test_concatenate(self):
        s = literal('foo') + 'bar'
        self.assertEqual(s.bits, (literal('foo'), 'bar'))

        s = 'foo' + literal('bar')
        self.assertEqual(s.bits, ('foo', literal('bar')))

        s = literal('foo') + literal('bar')
        self.assertEqual(s.bits, (literal('foobar'),))

        s = shell_literal('foo') + 'bar'
        self.assertEqual(s.bits, (shell_literal('foo'), 'bar'))

        s = 'foo' + shell_literal('bar')
        self.assertEqual(s.bits, ('foo', shell_literal('bar')))

        s = shell_literal('foo') + shell_literal('bar')
        self.assertEqual(s.bits, (shell_literal('foobar'),))

        s = literal('foo') + shell_literal('bar')
        self.assertEqual(s.bits, (literal('foo'), shell_literal('bar')))

        s = shell_literal('foo') + literal('bar')
        self.assertEqual(s.bits, (shell_literal('foo'), literal('bar')))


class TestJbos(TestCase):
    def test_construct_from_strings(self):
        s = jbos('foo', literal('bar'), shell_literal('baz'))
        self.assertEqual(s.bits, ('foo', literal('bar'), shell_literal('baz')))

    def test_construct_from_jbos(self):
        s = jbos(jbos('foo', literal('bar')), jbos(shell_literal('baz')))
        self.assertEqual(s.bits, ('foo', literal('bar'), shell_literal('baz')))

    def test_canonicalize(self):
        s = jbos('foo', 'bar')
        self.assertEqual(s.bits, ('foobar',))

        s = jbos(literal('foo'), literal('bar'))
        self.assertEqual(s.bits, (literal('foobar'),))

        s = jbos(shell_literal('foo'), shell_literal('bar'))
        self.assertEqual(s.bits, (shell_literal('foobar'),))

    def test_construct_invalid(self):
        self.assertRaises(TypeError, jbos, 123)

    def test_concatenate(self):
        s = jbos('foo') + literal('bar')
        self.assertEqual(s.bits, ('foo', literal('bar')))

        s = jbos('foo') + shell_literal('bar')
        self.assertEqual(s.bits, ('foo', shell_literal('bar')))

        s = jbos('foo') + 'bar'
        self.assertEqual(s.bits, ('foobar',))

        s = literal('foo') + jbos('bar')
        self.assertEqual(s.bits, (literal('foo'), 'bar'))

        s = shell_literal('foo') + jbos('bar')
        self.assertEqual(s.bits, (shell_literal('foo'), 'bar'))

        s = 'foo' + jbos('bar')
        self.assertEqual(s.bits, ('foobar',))

    def test_simplify(self):
        self.assertEqual(jbos().simplify(), '')

        self.assertEqual(jbos('foo').simplify(), 'foo')
        self.assertEqual(jbos(literal('foo')).simplify(), literal('foo'))
        self.assertEqual(jbos(shell_literal('foo')).simplify(),
                         shell_literal('foo'))

        self.assertEqual(jbos('foo', literal('bar')).simplify(),
                         jbos('foo', literal('bar')))

    def test_equality(self):
        self.assertTrue(jbos() == jbos())
        self.assertFalse(jbos() != jbos())

        self.assertTrue(jbos('foo') == jbos('foo'))
        self.assertTrue(jbos('foo', literal('bar')) ==
                        jbos('foo', literal('bar')))
        self.assertFalse(jbos('foo') != jbos('foo'))
        self.assertFalse(jbos('foo', literal('bar')) !=
                         jbos('foo', literal('bar')))

        self.assertFalse(jbos('foo', literal('bar')) ==
                         jbos('foo', literal('quux')))
        self.assertTrue(jbos('foo', literal('bar')) !=
                        jbos('foo', literal('quux')))

        self.assertFalse(jbos('foo') == jbos('foo', literal('bar')))
        self.assertTrue(jbos('foo') != jbos('foo', literal('bar')))


class TestJoin(TestCase):
    def test_join_empty(self):
        s = safe_str.join([], ',')
        self.assertEqual(s, '')

        s = safe_str.join([], literal(','))
        self.assertEqual(s, '')

    def test_join_strings(self):
        s = safe_str.join(['foo'], ',')
        self.assertEqual(s, 'foo')

        s = safe_str.join(['foo', 'bar'], ',')
        self.assertEqual(s, 'foo,bar')

    def test_join_literals(self):
        s = safe_str.join([literal('foo'), 'bar'], ',')
        self.assertEqual(s.bits, (literal('foo'), ',bar'))

        s = safe_str.join([shell_literal('foo'), 'bar'], ',')
        self.assertEqual(s.bits, (shell_literal('foo'), ',bar'))

        s = safe_str.join([literal('foo'), 'bar'], literal(','))
        self.assertEqual(s.bits, (literal('foo,'), 'bar'))

        s = safe_str.join([shell_literal('foo'), 'bar'], shell_literal(','))
        self.assertEqual(s.bits, (shell_literal('foo,'), 'bar'))


class TestSafeFormat(TestCase):
    def test_simple(self):
        self.assertEqual(safe_str.safe_format('foo'), 'foo')

    def test_auto(self):
        self.assertEqual(safe_str.safe_format('{}', 'foo'), 'foo')
        self.assertEqual(safe_str.safe_format('a{}z', 'foo'), 'afooz')

        foo = literal('foo')
        self.assertEqual(safe_str.safe_format('{}', foo), foo)
        self.assertEqual(safe_str.safe_format('a{}z', foo),
                         jbos('a', foo, 'z'))

        self.assertEqual(safe_str.safe_format('{}', MyString()), 'foo')
        self.assertEqual(safe_str.safe_format('a{}z', MyString()), 'afooz')

        self.assertEqual(safe_str.safe_format('{}', MyLiteral()), foo)
        self.assertEqual(safe_str.safe_format('a{}z', MyLiteral()),
                         jbos('a', foo, 'z'))

    def test_index(self):
        self.assertEqual(safe_str.safe_format('{0}', 'foo'), 'foo')
        self.assertEqual(safe_str.safe_format('a{0}z', 'foo'), 'afooz')

        foo = literal('foo')
        self.assertEqual(safe_str.safe_format('{0}', foo), foo)
        self.assertEqual(safe_str.safe_format('a{0}z', foo),
                         jbos('a', foo, 'z'))

        self.assertEqual(safe_str.safe_format('{0}', MyString()), 'foo')
        self.assertEqual(safe_str.safe_format('a{0}z', MyString()), 'afooz')

        self.assertEqual(safe_str.safe_format('{0}', MyLiteral()), foo)
        self.assertEqual(safe_str.safe_format('a{0}z', MyLiteral()),
                         jbos('a', foo, 'z'))

    def test_name(self):
        self.assertEqual(safe_str.safe_format('{f}', f='foo'), 'foo')
        self.assertEqual(safe_str.safe_format('a{f}z', f='foo'), 'afooz')

        foo = literal('foo')
        self.assertEqual(safe_str.safe_format('{f}', f=foo), foo)
        self.assertEqual(safe_str.safe_format('a{f}z', f=foo),
                         jbos('a', foo, 'z'))

        self.assertEqual(safe_str.safe_format('{f}', f=MyString()), 'foo')
        self.assertEqual(safe_str.safe_format('a{f}z', f=MyString()), 'afooz')

        self.assertEqual(safe_str.safe_format('{f}', f=MyLiteral()), foo)
        self.assertEqual(safe_str.safe_format('a{f}z', f=MyLiteral()),
                         jbos('a', foo, 'z'))

    def test_invalid(self):
        self.assertRaises(ValueError, safe_str.safe_format, '{}{0}', 'foo')
        self.assertRaises(ValueError, safe_str.safe_format, '{0}{}', 'foo')



import warnings
from six import itervalues

from . import builtin
from .. import path
from .. import shell
from ..backends.make import writer as make
from ..backends.ninja import writer as ninja
from ..build_inputs import build_input
from ..file_types import Directory, File, file_install_path, installify
from ..iterutils import flatten, iterate, unlistify


@build_input('install')
class InstallOutputs(object):
    def __init__(self, build_inputs, env):
        self.explicit = []
        self.implicit = []

    def add(self, item):
        if item not in self.explicit:
            self.explicit.append(item)

        for i in item.all:
            self._add_implicit(i)

    def _add_implicit(self, item):
        if not isinstance(item, File):
            raise TypeError('expected a file or directory')
        if item.path.root not in (path.Root.srcdir, path.Root.builddir):
            raise ValueError('external files are not installable')

        if item not in self.implicit:
            self.implicit.append(item)

        for i in item.install_deps:
            self._add_implicit(i)

    def __nonzero__(self):
        return self.__bool__()

    def __bool__(self):
        return bool(self.implicit)

    def __iter__(self):
        return iter(self.implicit)


def can_install(env):
    return all(i is not None for i in itervalues(env.install_dirs))


@builtin.function('builtins', 'build_inputs', 'env')
def install(builtins, build, env, *args):
    if len(args) == 0:
        return

    can_inst = can_install(env)
    if not can_inst:
        warnings.warn('unset installation directories; installation of this ' +
                      'build disabled')

    for i in args:
        if can_inst:
            build['install'].add(i)
        builtins['default'](i)

    return unlistify(tuple(installify(i, env) for i in args))


def _doppel_cmd(env, buildfile):
    doppel = env.tool('doppel')

    def wrapper(kind):
        cmd = buildfile.cmd_var(doppel)
        basename = cmd.name

        if kind != 'program':
            kind = 'data'
            cmd = [cmd] + doppel.data_args
        if basename.isupper():
            kind = kind.upper()

        name = '{name}_{kind}'.format(name=basename, kind=kind)
        cmd = buildfile.variable(name, cmd, buildfile.Section.command, True)
        return lambda *args, **kwargs: doppel(*args, cmd=cmd, **kwargs)
    return wrapper


def _install_commands(install_outputs, doppel):
    def install_line(output):
        cmd = doppel(output.install_kind)
        if isinstance(output, Directory):
            if output.files is not None:
                src = [i.path.relpath(output.path) for i in output.files]
                dst = file_install_path(output)
                return cmd('into', src, dst, directory=output.path)

            warnings.warn(
                ('installed directory {!r} has no matching files; did you ' +
                 'forget to set `include`?').format(output.path)
            )

        src = output.path
        dst = file_install_path(output)
        return cmd('onto', src, dst)

    return ([install_line(i) for i in install_outputs] +
            [i.post_install for i in install_outputs if i.post_install])


def _uninstall_command(install_outputs, rm):
    def uninstall_line(output):
        if isinstance(output, Directory):
            dst = file_install_path(output)
            return [dst.append(i.path.relpath(output.path)) for i in
                    iterate(output.files)]
        return [file_install_path(output)]

    return rm(flatten(uninstall_line(i) for i in install_outputs))


@make.post_rule
def make_install_rule(build_inputs, buildfile, env):
    install_outputs = build_inputs['install']
    if not install_outputs:
        return

    for i in path.InstallRoot:
        buildfile.variable(make.path_vars[i], env.install_dirs[i],
                           make.Section.path)
    if path.DestDir.destdir in make.path_vars:
        buildfile.variable(make.path_vars[path.DestDir.destdir],
                           env.variables.get('DESTDIR', ''), make.Section.path)

    buildfile.rule(
        target='install',
        deps='all',
        recipe=_install_commands(install_outputs, _doppel_cmd(env, buildfile)),
        phony=True
    )
    buildfile.rule(
        target='uninstall',
        recipe=[_uninstall_command(install_outputs, env.tool('rm'))],
        phony=True
    )


@ninja.post_rule
def ninja_install_rule(build_inputs, buildfile, env):
    install_outputs = build_inputs['install']
    if not install_outputs:
        return

    for i in path.InstallRoot:
        buildfile.variable(ninja.path_vars[i], env.install_dirs[i],
                           ninja.Section.path)
    if path.DestDir.destdir in ninja.path_vars:
        buildfile.variable(ninja.path_vars[path.DestDir.destdir],
                           env.variables.get('DESTDIR', ''),
                           ninja.Section.path)

    install = _install_commands(install_outputs, _doppel_cmd(env, buildfile))
    ninja.command_build(
        buildfile, env,
        output='install',
        inputs=['all'],
        command=shell.join_lines(install),
        phony=True
    )
    ninja.command_build(
        buildfile, env,
        output='uninstall',
        command=_uninstall_command(install_outputs, env.tool('rm')),
        phony=True
    )



import importlib
import pkgutil

from ..objutils import memoize

_builders = {}
_tools = {}
_tool_runners = {}


@memoize
def init():
    # Import all the packages in this directory so their hooks get run.
    for _, name, _ in pkgutil.walk_packages(__path__, '.'):
        importlib.import_module(name, __package__)


def builder(*args):
    if len(args) == 0:  # pragma: no cover
        raise TypeError('must provide at least one language')
    multi = len(args) > 1

    def wrapper(fn):
        for i in args:
            _builders[i] = (fn, multi)
        return fn
    return wrapper


def get_builder(env, lang):
    try:
        fn, multi = _builders[lang]
    except KeyError:
        raise ValueError('unknown language {!r}'.format(lang))
    return fn(env, lang) if multi else fn(env)


def tool(name, lang=None):
    def wrapper(fn):
        _tools[name] = fn
        if lang:
            _tool_runners[lang] = name
        return fn
    return wrapper


def get_tool(env, name):
    try:
        return _tools[name](env)
    except KeyError:
        raise ValueError('unknown tool {!r}'.format(name))


def get_tool_runner(lang):
    try:
        return _tool_runners[lang]
    except KeyError:
        raise ValueError('unknown tool runner {!r}'.format(lang))



import errno
import mock
import os
from collections import namedtuple

from . import *

from bfg9000 import path
from bfg9000.platforms import target
from bfg9000.safe_str import jbos

MockEnv = namedtuple('MockEnv', ['target_platform'])

path_variables = {
    path.Root.srcdir: '$(srcdir)',
    path.Root.builddir: None,
    path.InstallRoot.prefix: '$(prefix)',
    path.InstallRoot.bindir: '$(bindir)',
}


class TestPath(PathTestCase):
    def test_construct(self):
        p = self.Path('foo/bar', path.Root.srcdir)
        self.assertEqual(p.suffix, 'foo/bar')
        self.assertEqual(p.root, path.Root.srcdir)
        self.assertEqual(p.destdir, False)

        p = self.Path(r'foo\bar', path.Root.srcdir)
        self.assertEqual(p.suffix, 'foo/bar')
        self.assertEqual(p.root, path.Root.srcdir)
        self.assertEqual(p.destdir, False)

        p = self.Path('foo/.././bar', path.Root.srcdir)
        self.assertEqual(p.suffix, 'bar')
        self.assertEqual(p.root, path.Root.srcdir)
        self.assertEqual(p.destdir, False)

        p = self.Path('.', path.Root.srcdir)
        self.assertEqual(p.suffix, '')
        self.assertEqual(p.root, path.Root.srcdir)
        self.assertEqual(p.destdir, False)

        p = self.Path('foo/../.', path.Root.srcdir)
        self.assertEqual(p.suffix, '')
        self.assertEqual(p.root, path.Root.srcdir)
        self.assertEqual(p.destdir, False)

    def test_construct_absolute(self):
        p = self.Path('/foo/bar', path.Root.absolute)
        self.assertEqual(p.suffix, '/foo/bar')
        self.assertEqual(p.root, path.Root.absolute)
        self.assertEqual(p.destdir, False)

        p = self.Path('/foo/bar', path.Root.srcdir)
        self.assertEqual(p.suffix, '/foo/bar')
        self.assertEqual(p.root, path.Root.absolute)
        self.assertEqual(p.destdir, False)

        p = self.Path('C:/foo/bar', path.Root.srcdir)
        self.assertEqual(p.suffix, 'C:/foo/bar')
        self.assertEqual(p.root, path.Root.absolute)
        self.assertEqual(p.destdir, False)

        p = self.Path(r'C:\foo\bar', path.Root.srcdir)
        self.assertEqual(p.suffix, 'C:/foo/bar')
        self.assertEqual(p.root, path.Root.absolute)
        self.assertEqual(p.destdir, False)

        p = self.Path('C:/.', path.Root.srcdir)
        self.assertEqual(p.suffix, 'C:/')
        self.assertEqual(p.root, path.Root.absolute)
        self.assertEqual(p.destdir, False)

        p = self.Path('//server/mount/foo/bar', path.Root.srcdir)
        self.assertEqual(p.suffix, '//server/mount/foo/bar')
        self.assertEqual(p.root, path.Root.absolute)
        self.assertEqual(p.destdir, False)

        p = self.Path(r'\\server\mount\foo\bar', path.Root.srcdir)
        self.assertEqual(p.suffix, '//server/mount/foo/bar')
        self.assertEqual(p.root, path.Root.absolute)
        self.assertEqual(p.destdir, False)

        self.assertRaises(ValueError, self.Path, 'foo/bar', path.Root.absolute)
        self.assertRaises(ValueError, self.Path, 'c:foo')

    def test_construct_destdir(self):
        p = self.Path('foo/bar', path.InstallRoot.bindir, True)
        self.assertEqual(p.suffix, 'foo/bar')
        self.assertEqual(p.root, path.InstallRoot.bindir)
        self.assertEqual(p.destdir, True)

        self.assertRaises(ValueError, self.Path, 'foo/bar', path.Root.srcdir,
                          True)

    def test_ensure(self):
        self.assertEqual(self.Path.ensure('foo'), self.Path('foo'))
        self.assertEqual(self.Path.ensure('foo', path.Root.srcdir),
                         self.Path('foo', path.Root.srcdir))
        self.assertEqual(
            self.Path.ensure('foo', path.InstallRoot.bindir, True),
            self.Path('foo', path.InstallRoot.bindir, True)
        )

        p = self.Path('foo')
        self.assertIs(self.Path.ensure(p), p)
        self.assertIs(self.Path.ensure(p, path.Root.srcdir), p)
        self.assertIs(self.Path.ensure(p, path.InstallRoot.bindir, True), p)
        self.assertRaises(ValueError, self.Path.ensure, p, path.Root.srcdir,
                          strict=True)

    def test_equality(self):
        self.assertTrue(self.Path('a', path.Root.srcdir) ==
                        self.Path('a', path.Root.srcdir))
        self.assertFalse(self.Path('a', path.Root.srcdir) !=
                         self.Path('a', path.Root.srcdir))
        self.assertTrue(self.Path('a', path.InstallRoot.bindir, True) ==
                        self.Path('a', path.InstallRoot.bindir, True))
        self.assertFalse(self.Path('a', path.InstallRoot.bindir, True) !=
                         self.Path('a', path.InstallRoot.bindir, True))

        self.assertFalse(self.Path('a', path.Root.srcdir) ==
                         self.Path('a', path.Root.builddir))
        self.assertTrue(self.Path('a', path.Root.srcdir) !=
                        self.Path('a', path.Root.builddir))
        self.assertFalse(self.Path('a', path.Root.srcdir) ==
                         self.Path('b', path.Root.srcdir))
        self.assertTrue(self.Path('a', path.Root.srcdir) !=
                        self.Path('b', path.Root.srcdir))
        self.assertFalse(self.Path('a', path.InstallRoot.bindir, True) ==
                         self.Path('a', path.InstallRoot.bindir, False))
        self.assertTrue(self.Path('a', path.InstallRoot.bindir, True) !=
                        self.Path('a', path.InstallRoot.bindir, False))

        winpath = target.platform_info('winnt').Path
        linuxpath = target.platform_info('linux').Path
        self.assertFalse(winpath('a') == linuxpath('a'))
        self.assertTrue(winpath('a') != linuxpath('a'))

        self.assertFalse(self.Path('a', path.Root.srcdir) == 'a')
        self.assertTrue(self.Path('a', path.Root.srcdir) != 'a')

    def test_cross(self):
        for name in ('winnt', 'linux'):
            platform = target.platform_info(name)
            env = MockEnv(platform)

            p = self.Path('foo/bar', path.Root.srcdir)
            self.assertEqual(
                p.cross(env), platform.Path('foo/bar', path.Root.srcdir)
            )

    def test_parent(self):
        p = self.Path('foo/bar', path.Root.srcdir)
        self.assertEqual(p.parent(), self.Path('foo', path.Root.srcdir))

        p = self.Path('bar', path.Root.srcdir)
        self.assertEqual(p.parent(), self.Path('', path.Root.srcdir))

        p = self.Path('', path.Root.srcdir)
        self.assertRaises(ValueError, p.parent)

        p = self.Path('foo/bar', path.InstallRoot.bindir, True)
        self.assertEqual(
            p.parent(), self.Path('foo', path.InstallRoot.bindir, True)
        )

    def test_append(self):
        Root = path.Root
        p = self.Path('foo', Root.srcdir)
        self.assertEqual(p.append('bar'),
                         self.Path('foo/bar', Root.srcdir))
        self.assertEqual(p.append('.'), self.Path('foo', Root.srcdir))
        self.assertEqual(p.append('..'), self.Path('', Root.srcdir))
        self.assertEqual(p.append('../bar'), self.Path('bar', Root.srcdir))
        self.assertEqual(p.append(r'..\bar'), self.Path('bar', Root.srcdir))

        self.assertEqual(p.append('/bar'), self.Path('/bar', Root.absolute))
        self.assertEqual(p.append('C:/bar'),
                         self.Path('C:/bar', Root.absolute))
        self.assertEqual(p.append(r'C:\bar'),
                         self.Path('C:/bar', Root.absolute))
        self.assertEqual(p.append('//server/mount/bar'),
                         self.Path('//server/mount/bar', Root.absolute))
        self.assertEqual(p.append(r'\\server\mount\bar'),
                         self.Path('//server/mount/bar', Root.absolute))

        p = self.Path('foo', path.InstallRoot.bindir, True)
        self.assertEqual(p.append('bar'),
                         self.Path('foo/bar', path.InstallRoot.bindir, True))

        p = self.Path('foo', Root.srcdir)

    def test_ext(self):
        p = self.Path('foo.txt', path.Root.srcdir)
        self.assertEqual(p.ext(), '.txt')

    def test_addext(self):
        p = self.Path('foo', path.Root.srcdir)
        self.assertEqual(p.addext('.txt'),
                         self.Path('foo.txt', path.Root.srcdir))

        p = self.Path('foo', path.InstallRoot.bindir, True)
        self.assertEqual(p.addext('.txt'),
                         self.Path('foo.txt', path.InstallRoot.bindir, True))

    def test_stripext(self):
        p = self.Path('foo.txt', path.Root.srcdir)
        self.assertEqual(p.stripext(), self.Path('foo', path.Root.srcdir))
        p = self.Path('foo.txt', path.Root.srcdir)
        self.assertEqual(p.stripext('.cpp'),
                         self.Path('foo.cpp', path.Root.srcdir))

        p = self.Path('foo', path.Root.srcdir)
        self.assertEqual(p.stripext(), self.Path('foo', path.Root.srcdir))
        p = self.Path('foo', path.Root.srcdir)
        self.assertEqual(p.stripext('.cpp'),
                         self.Path('foo.cpp', path.Root.srcdir))

        p = self.Path('foo.txt', path.InstallRoot.bindir, True)
        self.assertEqual(p.stripext(),
                         self.Path('foo', path.InstallRoot.bindir, True))

    def test_splitleaf(self):
        p = self.Path('foo/bar/baz', path.Root.srcdir)
        self.assertEqual(p.splitleaf(),
                         (self.Path('foo/bar', path.Root.srcdir), 'baz'))

    def test_split(self):
        p = self.Path('foo/bar/baz', path.Root.srcdir)
        self.assertEqual(p.split(), ['foo', 'bar', 'baz'])

    def test_basename(self):
        p = self.Path('foo/bar', path.Root.srcdir)
        self.assertEqual(p.basename(), 'bar')

    def test_relpath_relative(self):
        p = self.Path('foo/bar', path.Root.srcdir)
        self.assertEqual(p.relpath(p), '.')
        self.assertEqual(p.relpath(p, 'pre'), 'pre')

        self.assertEqual(p.relpath(self.Path('foo', path.Root.srcdir)), 'bar')
        self.assertEqual(p.relpath(self.Path('foo', path.Root.srcdir), 'pre'),
                         self.ospath.join('pre', 'bar'))

        self.assertEqual(p.relpath(self.Path('baz', path.Root.srcdir)),
                         self.ospath.join('..', 'foo', 'bar'))
        self.assertEqual(p.relpath(self.Path('baz', path.Root.srcdir), 'pre'),
                         self.ospath.join('pre', '..', 'foo', 'bar'))

        self.assertEqual(p.relpath(self.Path('.', path.Root.srcdir)),
                         self.ospath.join('foo', 'bar'))
        self.assertEqual(p.relpath(self.Path('.', path.Root.srcdir), 'pre'),
                         self.ospath.join('pre', 'foo', 'bar'))

        p = self.Path('.', path.Root.srcdir)
        self.assertEqual(p.relpath(path.Path('foo', path.Root.srcdir)), '..')
        self.assertEqual(p.relpath(path.Path('foo', path.Root.srcdir), 'pre'),
                         self.ospath.join('pre', '..'))

        p = self.Path('foo/bar', path.Root.srcdir)
        self.assertRaises(
            ValueError, lambda: p.relpath(self.Path('foo', path.Root.builddir))
        )

    def test_relpath_absolute(self):
        p = self.Path('/foo/bar', path.Root.srcdir)
        self.assertEqual(
            p.relpath(self.Path('start', path.Root.srcdir)),
            self.ospath.join(self.ospath.sep, 'foo', 'bar')
        )
        self.assertEqual(
            p.relpath(self.Path('start', path.Root.srcdir), 'pre'),
            self.ospath.join(self.ospath.sep, 'foo', 'bar')
        )

        self.assertEqual(
            p.relpath(self.Path('/start', path.Root.srcdir)),
            self.ospath.join(self.ospath.sep, 'foo', 'bar')
        )
        self.assertEqual(
            p.relpath(self.Path('/start', path.Root.srcdir), 'pre'),
            self.ospath.join(self.ospath.sep, 'foo', 'bar')
        )

        p = self.Path(r'C:\foo\bar', path.Root.srcdir)
        self.assertEqual(
            p.relpath(self.Path(r'C:\start', path.Root.srcdir)),
            'C:' + self.ospath.sep + self.ospath.join('foo', 'bar')
        )
        self.assertEqual(
            p.relpath(self.Path(r'C:\start', path.Root.srcdir), 'pre'),
            'C:' + self.ospath.sep + self.ospath.join('foo', 'bar')
        )

    def test_reroot(self):
        p = self.Path('foo/bar', path.Root.srcdir)
        self.assertEqual(p.reroot(path.Root.builddir),
                         self.Path('foo/bar', path.Root.builddir))

        p = self.Path('foo/bar', path.InstallRoot.bindir, True)
        self.assertEqual(p.reroot(path.InstallRoot.libdir),
                         self.Path('foo/bar', path.InstallRoot.libdir, True))

    def test_to_json(self):
        p = self.Path('foo', path.Root.srcdir)
        self.assertEqual(p.to_json(), ('foo', 'srcdir', False))

        p = self.Path('foo', path.InstallRoot.bindir, True)
        self.assertEqual(p.to_json(), ('foo', 'bindir', True))

    def test_from_json(self):
        p = self.Path.from_json(['foo', 'srcdir', False])
        self.assertEqual(p, self.Path('foo', path.Root.srcdir))

        p = self.Path.from_json(['foo', 'bindir', True])
        self.assertEqual(p, self.Path('foo', path.InstallRoot.bindir, True))

    def test_realize_srcdir(self):
        p = self.Path('foo', path.Root.srcdir)
        self.assertEqual(p.realize(path_variables),
                         self.ospath.join('$(srcdir)', 'foo'))
        self.assertEqual(p.realize(path_variables, executable=True),
                         self.ospath.join('$(srcdir)', 'foo'))

        p = self.Path('foo/bar', path.Root.srcdir)
        self.assertEqual(p.realize(path_variables),
                         self.ospath.join('$(srcdir)', 'foo', 'bar'))
        self.assertEqual(p.realize(path_variables, executable=True),
                         self.ospath.join('$(srcdir)', 'foo', 'bar'))

    def test_realize_builddir(self):
        p = self.Path('foo', path.Root.builddir)
        self.assertEqual(p.realize(path_variables), 'foo')
        self.assertEqual(p.realize(path_variables, executable=True),
                         self.ospath.join('.', 'foo'))

        p = self.Path('foo/bar', path.Root.builddir)
        self.assertEqual(p.realize(path_variables),
                         self.ospath.join('foo', 'bar'))
        self.assertEqual(p.realize(path_variables, executable=True),
                         self.ospath.join('foo', 'bar'))

    def test_realize_absolute(self):
        ospath = self.ospath
        p = self.Path('/foo/bar', path.Root.builddir)
        self.assertEqual(p.realize(path_variables),
                         ospath.join(ospath.sep, 'foo', 'bar'))
        self.assertEqual(p.realize(path_variables, executable=True),
                         ospath.join(ospath.sep, 'foo', 'bar'))

        p = self.Path(r'C:\foo\bar', path.Root.builddir)
        self.assertEqual(p.realize(path_variables),
                         'C:' + ospath.sep + ospath.join('foo', 'bar'))
        self.assertEqual(p.realize(path_variables, executable=True),
                         'C:' + ospath.sep + ospath.join('foo', 'bar'))

    def test_realize_srcdir_empty(self):
        p = self.Path('', path.Root.srcdir)
        self.assertEqual(p.realize(path_variables), '$(srcdir)')
        self.assertEqual(p.realize(path_variables, executable=True),
                         '$(srcdir)')

    def test_realize_builddir_empty(self):
        p = self.Path('', path.Root.builddir)
        self.assertEqual(p.realize(path_variables), '.')
        self.assertEqual(p.realize(path_variables, executable=True), '.')

    def test_realize_destdir(self):
        path_vars_with_destdir = path_variables.copy()
        path_vars_with_destdir[path.DestDir.destdir] = '$(destdir)'

        p = self.Path('foo', path.InstallRoot.bindir, True)
        self.assertEqual(p.realize(path_variables),
                         self.ospath.join('$(bindir)', 'foo'))
        self.assertEqual(p.realize(path_vars_with_destdir),
                         self.ospath.join('$(destdir)$(bindir)', 'foo'))

    def test_string(self):
        ospath = self.ospath
        paths = {path.Root.srcdir: self.Path('/srcdir', path.Root.absolute)}

        p = self.Path('/foo/bar', path.Root.absolute)
        self.assertEqual(p.string(), ospath.join(ospath.sep, 'foo', 'bar'))

        p = self.Path('foo/bar', path.Root.srcdir)
        self.assertEqual(
            p.string(paths), ospath.join(ospath.sep, 'srcdir', 'foo', 'bar')
        )

        p = self.Path('.', path.Root.srcdir)
        self.assertEqual(p.string(paths), ospath.join(ospath.sep, 'srcdir'))

    def test_hash(self):
        d = {self.Path('.', path.Root.srcdir),
             self.Path('.', path.Root.builddir),
             self.Path('foo', path.Root.srcdir),
             self.Path('bar', path.InstallRoot.bindir),
             self.Path('bar', path.InstallRoot.bindir, destdir=True)}
        self.assertEqual(len(d), 5)

    def test_add(self):
        p = self.Path('foo/bar', path.Root.srcdir)
        result = p + 'baz'
        self.assertEqual(type(result), jbos)
        self.assertEqual(result.bits, (p, 'baz'))

        result = 'baz' + p
        self.assertEqual(type(result), jbos)
        self.assertEqual(result.bits, ('baz', p))


class TestAbsPath(TestCase):
    def test_abspath(self):
        with mock.patch('os.getcwd', return_value=r'/base'):
            self.assertEqual(path.abspath('foo'),
                             path.Path('/base/foo', path.Root.absolute))
            self.assertEqual(path.abspath('/foo/bar'),
                             path.Path('/foo/bar', path.Root.absolute))

    def test_drive(self):
        with mock.patch('os.getcwd', return_value=r'C:\base'):
            self.assertEqual(path.abspath('foo'),
                             path.Path('C:/base/foo', path.Root.absolute))
            self.assertEqual(path.abspath('/foo/bar'),
                             path.Path('C:/foo/bar', path.Root.absolute))
            self.assertEqual(path.abspath('D:/foo/bar'),
                             path.Path('D:/foo/bar', path.Root.absolute))


class TestInstallPath(TestCase):
    def test_install_path_file(self):
        p = path.Path('foo/bar', path.Root.srcdir)
        self.assertEqual(path.install_path(p, path.InstallRoot.bindir),
                         path.Path('bar', path.InstallRoot.bindir, True))

        p = path.Path('foo/bar', path.Root.builddir)
        self.assertEqual(path.install_path(p, path.InstallRoot.bindir),
                         path.Path('foo/bar', path.InstallRoot.bindir, True))

        p = path.abspath('/foo/bar')
        with self.assertRaises(TypeError):
            path.install_path(p, path.InstallRoot.bindir)

    def test_install_path_directory(self):
        p = path.Path('foo/bar', path.Root.srcdir)
        self.assertEqual(path.install_path(p, path.InstallRoot.bindir, True),
                         path.Path('', path.InstallRoot.bindir, True))

        p = path.Path('foo/bar', path.Root.builddir)
        self.assertEqual(path.install_path(p, path.InstallRoot.bindir, True),
                         path.Path('foo/bar', path.InstallRoot.bindir, True))

        p = path.abspath('/foo/bar')
        with self.assertRaises(TypeError):
            path.install_path(p, path.InstallRoot.bindir, True)

    def test_install_path_cross(self):
        for name in ('winnt', 'linux'):
            platform = target.platform_info(name)
            env = MockEnv(platform)

            p = path.Path('foo/bar', path.Root.srcdir)
            self.assertEqual(
                path.install_path(p, path.InstallRoot.bindir, cross=env),
                platform.Path('bar', path.InstallRoot.bindir)
            )

            p = path.Path('foo/bar', path.Root.builddir)
            self.assertEqual(
                path.install_path(p, path.InstallRoot.bindir, cross=env),
                platform.Path('foo/bar', path.InstallRoot.bindir)
            )

            p = path.Path('/foo/bar', path.Root.absolute)
            self.assertEqual(
                path.install_path(p, path.InstallRoot.bindir, cross=env),
                platform.Path('/foo/bar', path.Root.absolute)
            )

    def test_install_path_cross_directory(self):
        for name in ('winnt', 'linux'):
            platform = target.platform_info(name)
            env = MockEnv(platform)

            p = path.Path('foo/bar', path.Root.srcdir)
            self.assertEqual(path.install_path(p, path.InstallRoot.bindir,
                                               True, cross=env),
                             platform.Path('', path.InstallRoot.bindir))

            p = path.Path('foo/bar', path.Root.builddir)
            self.assertEqual(path.install_path(p, path.InstallRoot.bindir,
                                               True, cross=env),
                             platform.Path('foo/bar', path.InstallRoot.bindir))

            p = path.Path('/foo/bar', path.Root.absolute)
            self.assertEqual(path.install_path(p, path.InstallRoot.bindir,
                                               True, cross=env),
                             platform.Path('/foo/bar', path.Root.absolute))


class TestCommonPrefix(TestCase):
    def test_empty(self):
        self.assertEqual(path.commonprefix([]), None)

    def test_single(self):
        p = path.Path('foo/bar')
        self.assertEqual(path.commonprefix([p]), p)

    def test_multi_same(self):
        p = path.Path('foo/bar')
        self.assertEqual(path.commonprefix([p, p]), p)

    def test_multi_partial_match(self):
        p = path.Path('foo/bar')
        q = path.Path('foo/baz')
        self.assertEqual(path.commonprefix([p, q]), p.parent())

    def test_multi_subset(self):
        p = path.Path('foo/bar')
        q = path.Path('foo/bar/baz')
        self.assertEqual(path.commonprefix([p, q]), p)

    def test_multi_no_match(self):
        p = path.Path('foo/bar')
        q = path.Path('baz/quux')
        self.assertEqual(path.commonprefix([p, q]), path.Path(''))


class TestWrappedOsPath(TestCase):
    def test_wrap(self):
        mocked = mock.MagicMock(return_value=True)
        mocked.__name__ = 'foo'
        f = path._wrap_ospath(mocked)
        f(path.Path('/foo/bar'))
        mocked.assert_called_once_with(os.path.join(os.path.sep, 'foo', 'bar'))


class TestSamefile(TestCase):
    def test_real(self):
        with mock.patch('os.path.samefile', lambda x, y: x == y, create=True):
            self.assertEqual(path.samefile(path.Path('/foo/bar'),
                                           path.Path('/foo/bar')), True)

    def test_polyfill(self):
        class OsPath(object):
            def __init__(self):
                for i in ('isabs', 'normpath', 'realpath', 'expanduser'):
                    setattr(self, i, getattr(os.path, i))

        with mock.patch('os.path', OsPath()):
            self.assertEqual(path.samefile(path.Path('/foo/bar'),
                                           path.Path('/foo/bar')), True)


class TestMakedirs(TestCase):
    def test_success(self):
        with mock.patch('os.makedirs') as os_makedirs:
            path.makedirs('foo')
            path.makedirs('bar', 0o666)
            self.assertEqual(os_makedirs.mock_calls, [
                mock.call('foo', 0o777),
                mock.call('bar', 0o666),
            ])

    def test_exists(self):
        def mock_makedirs(path, mode):
            raise OSError(errno.EEXIST, 'msg')

        with mock.patch('os.makedirs', mock_makedirs), \
             mock.patch('os.path.isdir', lambda x: x == 'dir'):  # noqa
            self.assertRaises(OSError, path.makedirs, 'file')
            self.assertRaises(OSError, path.makedirs, 'file', exist_ok=True)
            self.assertRaises(OSError, path.makedirs, 'dir')
            path.makedirs('dir', exist_ok=True)

    def test_other_error(self):
        def mock_makedirs(path, mode):
            raise OSError(errno.EPERM, 'msg')

        with mock.patch('os.makedirs', mock_makedirs):
            self.assertRaises(OSError, path.makedirs, 'file')


class TestPushd(TestCase):
    def test_basic(self):
        with mock.patch('os.getcwd', return_value='cwd'), \
             mock.patch('os.chdir') as os_chdir:  # noqa
            with path.pushd('foo'):
                self.assertEqual(os_chdir.mock_calls, [mock.call('foo')])
            self.assertEqual(os_chdir.mock_calls, [
                mock.call('foo'), mock.call('cwd')
            ])

    def test_makedirs(self):
        with mock.patch('os.makedirs') as os_makedirs, \
             mock.patch('os.getcwd', return_value='cwd'), \
             mock.patch('os.chdir') as os_chdir:  # noqa
            with path.pushd('foo', makedirs=True):
                self.assertEqual(os_makedirs.mock_calls, [
                    mock.call('foo', 0o777)
                ])
                self.assertEqual(os_chdir.mock_calls, [mock.call('foo')])
            self.assertEqual(os_chdir.mock_calls, [
                mock.call('foo'), mock.call('cwd')
            ])



from itertools import chain

from . import builder
from .. import options as opts, safe_str, shell
from .common import BuildCommand, Builder, choose_builder
from ..file_types import HeaderFile, SourceFile
from ..iterutils import first, iterate, listify
from ..languages import known_langs
from ..path import Path
from ..versioning import detect_version

with known_langs.make('yacc') as x:
    x.vars(compiler='YACC', flags='YFLAGS')
    x.exts(source=['.y'])

_posix_cmds = ['yacc', 'bison']
_windows_cmds = ['yacc', 'bison', 'win_bison']


@builder('yacc')
def yacc_builder(env):
    cmds = (_windows_cmds if env.host_platform.family == 'windows'
            else _posix_cmds)
    return choose_builder(env, known_langs['yacc'], cmds, (YaccBuilder,))


class YaccBuilder(Builder):
    def __init__(self, env, langinfo, command, version_output):
        Builder.__init__(self, langinfo.name,
                         *self._parse_brand(version_output))

        name = langinfo.var('compiler').lower()
        lflags_name = langinfo.var('flags').lower()
        lflags = shell.split(env.getvar(langinfo.var('flags'), ''))

        self.transpiler = YaccCompiler(self, env, name, name, command,
                                      flags=(lflags_name, lflags))

    @staticmethod
    def _parse_brand(version_output):
        if 'bison' in version_output:
            return 'bison', detect_version(version_output)
        return 'unknown', None

    @staticmethod
    def check_command(env, command):
        return env.execute(command + ['--version'], stdout=shell.Mode.pipe,
                           stderr=shell.Mode.devnull)


class YaccCompiler(BuildCommand):
    @property
    def deps_flavor(self):
        return None

    @property
    def needs_libs(self):
        return False

    @property
    def num_outputs(self):
        return 1

    def _call(self, cmd, input, output, flags=None):
        return list(chain(
            cmd, iterate(flags), [input, '-o', first(output)]
        ))

    def pre_build(self, build, name, context):
        name = listify(name)
        return opts.option_list(['--defines=' + name[1]]
                                if len(name) > 1 else [])

    def _output_lang(self, options):
        filtered = options.filter(opts.lang) if options else None
        if filtered:
            return filtered[-1].value
        return 'c'

    def default_name(self, input, context):
        options = getattr(context, 'user_options', None)
        lang = known_langs[self._output_lang(options)]
        return [
            input.path.stripext('.tab' + lang.default_ext('source')).suffix,
            input.path.stripext('.tab' + lang.default_ext('header')).suffix
        ]

    def output_file(self, name, context):
        name = listify(name)
        options = getattr(context, 'user_options', None)
        lang = self._output_lang(options)
        src = SourceFile(Path(name[0]), lang)

        if len(name) > 2:
            raise ValueError('too many output names')
        elif len(name) == 2:
            return [src, HeaderFile(Path(name[1]), lang)]
        return src

    def flags(self, options, output=None, mode='normal'):
        flags = []
        for i in options:
            if isinstance(i, opts.define):
                if i.value:
                    flags.append('-D' + i.name + '=' + i.value)
                else:
                    flags.append('-D' + i.name)
            elif isinstance(i, opts.warning):
                for j in i.value:
                    if j == opts.WarningValue.disable:
                        flags.append('-w')
                    else:
                        raise ValueError('unsupported warning level {!r}'
                                         .format(j))
            elif isinstance(i, opts.lang):
                flags.append('--language=' + i.value)
            elif isinstance(i, safe_str.stringy_types):
                flags.append(i)
            else:
                raise TypeError('unknown option type {!r}'.format(type(i)))
        return flags



import errno

from .arguments.parser import ArgumentParser
from .builtins import builtin, init as builtin_init
from .build_inputs import BuildInputs
from .path import exists, Path, pushd, Root
from .iterutils import listify
from .tools import init as tools_init

bfgfile = 'build.bfg'
optsfile = 'options.bfg'

user_description = """
These arguments are defined by the options.bfg file in the project's source
directory. To disambiguate them from built-in arguments, you may prefix the
argument name with `-x`. For example, `--foo` may also be written as `--x-foo`.
"""


def is_srcdir(path):
    return exists(path.append(bfgfile))


def _execute_file(f, filename, builtin_dict):
    code = compile(f.read(), filename, 'exec')
    try:
        exec(code, builtin_dict)
    except SystemExit:
        pass


def load_toolchain(env, filename, reload=False):
    builtin_init()
    tools_init()
    if reload:
        env.init_variables()

    builtin_dict = builtin.toolchain.bind(env=env, reload=reload)
    with open(filename.string(), 'r') as f:
        _execute_file(f, f.name, builtin_dict)

    if not reload:
        env.toolchain.path = filename


def _execute_options(env, optspath, parent=None, usage='parse'):
    prog = parent.prog if parent else optspath.basename()
    parser = ArgumentParser(prog=prog, parents=listify(parent),
                            add_help=False)
    executed = False

    try:
        with open(optspath.string(env.base_dirs), 'r') as f, \
             pushd(env.srcdir.string()):  # noqa
            group = parser.add_argument_group('project-defined arguments',
                                              description=user_description)
            group.usage = usage

            builtin_dict = builtin.options.bind(env=env, parser=group)
            _execute_file(f, optspath.basename(), builtin_dict)
            builtin.options.run_post(builtin_dict, env=env, parser=group)
        executed = True
    except IOError as e:
        if e.errno != errno.ENOENT:
            raise

    return parser, executed


def _execute_configure(env, argv, bfgpath, extra_bootstrap=[]):
    build = BuildInputs(env, bfgpath, extra_bootstrap)
    builtin_dict = builtin.build.bind(build_inputs=build, argv=argv, env=env)

    with open(bfgpath.string(env.base_dirs), 'r') as f, \
         pushd(env.srcdir.string()):  # noqa
        _execute_file(f, bfgpath.basename(), builtin_dict)
        builtin.build.run_post(builtin_dict, build_inputs=build, argv=argv,
                               env=env)
    return build


def fill_user_help(env, parent, filename=optsfile):
    builtin_init()
    optspath = Path(filename, Root.srcdir)
    return _execute_options(env, optspath, parent, usage='help')[0]


def configure_build(env, bfgfile=bfgfile, optsfile=optsfile):
    builtin_init()
    bfgpath = Path(bfgfile, Root.srcdir)
    optspath = Path(optsfile, Root.srcdir)

    parser, executed = _execute_options(env, optspath)
    argv = parser.parse_args(env.extra_args)

    extra_bootstrap = [optspath] if executed else []
    return _execute_configure(env, argv, bfgpath, extra_bootstrap)



import os

from . import builtin
from .file_types import FileList
from ..backends.make import writer as make
from ..backends.ninja import writer as ninja
from ..build_inputs import Edge
from ..file_types import File
from ..path import Path
from ..versioning import SpecifierSet


class CopyFile(Edge):
    __modes = {'copy', 'symlink', 'hardlink'}
    msbuild_output = True

    def __init__(self, build, env, output, file, mode='copy', extra_deps=None,
                 description=None):
        if mode not in self.__modes:
            raise ValueError('unrecognized copy mode {!r}'.format(mode))

        self.mode = mode
        self.copier = env.tool(mode)
        self.file = file
        Edge.__init__(self, build, output, None, extra_deps, description)

    @staticmethod
    def convert_args(builtins, name, file, kwargs):
        directory = kwargs.pop('directory', None)
        if directory:
            directory = Path.ensure(directory, strict=True)
        file = builtins['auto_file'](file)

        def pathfn(file):
            if name is None:
                path = file.path.reroot()
                if directory:
                    return directory.append(path.suffix)
                return path
            return Path(name)

        output = file.clone(pathfn)
        return output, file, kwargs


@builtin.function('builtins', 'build_inputs', 'env')
@builtin.type(File, short_circuit=False, first_optional=True)
def copy_file(builtins, build, env, name, file, **kwargs):
    # Note: this only handles single files. File objects with multiple
    # related files (e.g. a DLL and its import library) will only copy the
    # primary file. In practice, this shouldn't matter, as this function is
    # mostly useful for copying data files to the build directory.
    output, file, kwargs = CopyFile.convert_args(builtins, name, file, kwargs)
    return CopyFile(build, env, output, file, **kwargs).public_output


@builtin.function('builtins')
@builtin.type(FileList, in_type=object)
def copy_files(builtins, files, **kwargs):
    return FileList(builtins['copy_file'], files, **kwargs)


@make.rule_handler(CopyFile)
def make_copy_file(rule, build_inputs, buildfile, env):
    copier = rule.copier
    recipename = make.var('RULE_{}'.format(copier.rule_name.upper()))

    if hasattr(copier, 'transform_input'):
        input_var = make.qvar('1')
        args = [copier.transform_input(rule.file, rule.raw_output)]
    else:
        input_var = make.qvar('<')
        args = []

    if not buildfile.has_variable(recipename):
        buildfile.define(recipename, [copier(input_var, make.qvar('@'))])

    buildfile.rule(
        target=rule.output,
        deps=rule.file,
        recipe=make.Call(recipename, *args)
    )


@ninja.rule_handler(CopyFile)
def ninja_copy_file(rule, build_inputs, buildfile, env):
    copier = rule.copier

    variables = {}
    if rule.description:
        variables['description'] = rule.description

    if hasattr(copier, 'transform_input'):
        input_var = ninja.var('input')
        variables[input_var] = copier.transform_input(
            rule.file, rule.raw_output
        )
    else:
        input_var = ninja.var('in')

    if not buildfile.has_rule(copier.rule_name):
        desc = rule.mode + ' => ' + ninja.var('out')
        buildfile.rule(
            name=copier.rule_name,
            command=copier(input_var, ninja.var('out')),
            description=desc
        )

    buildfile.build(
        output=rule.output,
        rule=copier.rule_name,
        inputs=rule.file,
        variables=variables
    )


try:
    from ..backends.msbuild import writer as msbuild

    @msbuild.rule_handler(CopyFile)
    def msbuild_copy_file(rule, build_inputs, solution, env):
        copier = rule.copier
        if hasattr(copier, 'transform_input'):
            input = copier.transform_input(rule.file, rule.raw_output)
        else:
            input = rule.file

        extra_attrs = {}
        if rule.mode == 'symlink':
            if ( env.backend_version and env.backend_version in
                 SpecifierSet('>=15') ):
                extra_attrs['UseSymboliclinksIfPossible'] = 'true'
        elif rule.mode == 'hardlink':
            extra_attrs['UseHardlinksIfPossible'] = 'true'

        name = os.path.join('copy_file_tasks', rule.output[0].path.suffix)
        project = msbuild.CommandProject(
            env, name=name, commands=[msbuild.CommandProject.task(
                'Copy', SourceFiles=input, DestinationFiles=rule.raw_output,
                **extra_attrs
            )],
            dependencies=solution.dependencies(rule.extra_deps),
        )
        solution[rule.output[0]] = project
except ImportError:  # pragma: no cover
    pass



import re

from . import *


class TestEnvVars(IntegrationTest):
    def __init__(self, *args, **kwargs):
        IntegrationTest.__init__(self, 'env_vars', *args, **kwargs)

    @skip_if_backend('msbuild')
    def test_test(self):
        self.build('test')

    def test_command(self):
        assertRegex(self, self.build('script'),
                    re.compile(r'^\s*hello script$', re.MULTILINE))
        self.assertExists(output_file('file'))



from collections import Counter
from itertools import chain
from six import iteritems, itervalues, string_types

from . import builtin
from .file_types import make_immediate_file
from .install import can_install
from .. import options as opts, path
from ..build_inputs import build_input
from ..file_types import *
from ..iterutils import flatten, iterate, uniques, recursive_walk
from ..objutils import objectify
from ..packages import CommonPackage
from ..safe_str import literal, shell_literal
from ..shell import posix as pshell
from ..shell.syntax import Syntax, Writer
from ..tools.pkg_config import PkgConfigPackage
from ..versioning import simplify_specifiers, Specifier, SpecifierSet

build_input('pkg_config')(lambda build_inputs, env: [])


class Requirement(object):
    def __init__(self, name, version=None):
        self.name = name
        self.version = objectify(version or '', SpecifierSet)

    def __and__(self, rhs):
        result = Requirement(self.name, self.version)
        result &= rhs
        return result

    def __iand__(self, rhs):
        if self.name != rhs.name:
            raise ValueError('requirement names do not match')
        self.version = self.version & rhs.version
        return self

    def __eq__(self, rhs):
        return (type(self) == type(rhs) and self.name == rhs.name and
                self.version == rhs.version)

    def __ne__(self, rhs):
        return not (self == rhs)

    def split(self, single=False):
        specs = simplify_specifiers(self.version)
        if len(specs) == 0:
            return [SimpleRequirement(self.name)]
        if single and len(specs) > 1:
            raise ValueError(
                ("multiple specifiers ({}) used in pkg-config requirement " +
                 "for '{}'").format(self.version, self.name)
            )
        return [SimpleRequirement(self.name, i) for i in specs]

    def __hash__(self):
        return hash((self.name, self.version))

    def __repr__(self):
        return '<Requirement({!r})>'.format(self._string())

    def _string(self):  # pragma: no cover
        return self.name + str(self.version)


class SimpleRequirement(object):
    def __init__(self, name, version=None):
        self.name = name
        self.version = (None if version is None else
                        objectify(version, Specifier))

    def __eq__(self, rhs):
        return (type(self) == type(rhs) and self.name == rhs.name and
                self.version == rhs.version)

    def __ne__(self, rhs):
        return not (self == rhs)

    def _safe_str(self):
        if not self.version:
            return shell_literal(self.name)
        op = self.version.operator
        if op == '==':
            op = '='
        return shell_literal('{name} {op} {version}'.format(
            name=self.name, op=op, version=self.version.version
        ))

    def __hash__(self):
        return hash((self.name, self.version))

    def __repr__(self):
        return '<SimpleRequirement({!r})>'.format(self._string())

    def _string(self):  # pragma: no cover
        return self.name + str(self.version)


class RequirementSet(object):
    def __init__(self, iterable=None):
        self._reqs = {}
        if iterable:
            for i in iterable:
                self.add(i)

    def add(self, item):
        if item.name not in self._reqs:
            self._reqs[item.name] = item
        else:
            self._reqs[item.name] &= item

    def remove(self, name):
        del self._reqs[name]

    def update(self, other):
        for i in other:
            self.add(i)

    def merge_from(self, other):
        items = list(other)
        for i in items:
            if i.name in self._reqs:
                self._reqs[i.name] &= i
                other.remove(i.name)

    def split(self, single=False):
        return sorted(flatten(i.split(single) for i in self),
                      key=lambda x: x.name)

    def __iter__(self):
        return itervalues(self._reqs)

    def __repr__(self):
        return '<RequirementSet({!r})>'.format(
            [i._string() for i in iter(self)]
        )


class PkgConfigInfo(object):
    directory = path.Path('pkgconfig')

    class _simple_property(object):
        def __init__(self, fn):
            self.fn = fn

        def __get__(self, obj, objtype=None):
            if obj is None:
                return self
            return getattr(obj, '_' + self.fn.__name__)

        def __set__(self, obj, value):
            setattr(obj, '_' + self.fn.__name__, self.fn(obj, value))

    def __init__(self, builtins, build, name=None, desc_name=None, desc=None,
                 url=None, version=None, requires=None, requires_private=None,
                 conflicts=None, includes=None, libs=None, libs_private=None,
                 options=None, link_options=None, link_options_private=None,
                 lang=None, auto_fill=True):
        self._builtins = builtins
        self.auto_fill = auto_fill

        self.name = name
        self.desc_name = desc_name
        self.desc = desc
        self.url = url
        self.version = version
        self.lang = lang or build['project']['lang']

        self.requires = requires
        self.requires_private = requires_private
        self.conflicts = conflicts

        self.includes = includes
        self.libs = libs
        self.libs_private = libs_private
        self.options = pshell.listify(options, type=opts.option_list)
        self.link_options = pshell.listify(link_options, type=opts.option_list)
        self.link_options_private = pshell.listify(link_options_private,
                                                   type=opts.option_list)

    @property
    def output(self):
        return PkgConfigPcFile(self.directory.append(self.name + '.pc'))

    @_simple_property
    def includes(self, value):
        return uniques(self._builtins['header_directory'](i)
                       for i in iterate(value)) if value is not None else None

    @_simple_property
    def libs(self, value):
        return (uniques(self._library(i) for i in iterate(value))
                if value is not None else None)

    @_simple_property
    def libs_private(self, value):
        return (uniques(self._library(i) for i in iterate(value))
                if value is not None else None)

    @_simple_property
    def requires(self, value):
        return (self._filter_packages(iterate(value))
                if value is not None else None)

    @_simple_property
    def requires_private(self, value):
        return (self._filter_packages(iterate(value))
                if value is not None else None)

    @_simple_property
    def conflicts(self, value):
        return (self._filter_packages(iterate(value))[0]
                if value is not None else None)

    def _library(self, lib):
        if isinstance(lib, DualUseLibrary):
            return lib
        return self._builtins['library'](lib)

    def _write_variable(self, out, name, value):
        out.write(name, Syntax.variable)
        out.write_literal('=')
        out.write(value, Syntax.variable)
        out.write_literal('\n')

    def _write_field(self, out, name, value, syntax=Syntax.variable, **kwargs):
        if value:
            out.write(name, Syntax.variable)
            out.write_literal(': ')
            out.write_each(iterate(value), syntax, **kwargs)
            out.write_literal('\n')

    def write(self, out, env):
        out = Writer(out)

        data = self._process_inputs(env)

        for i in path.InstallRoot:
            if i != path.InstallRoot.bindir:
                self._write_variable(out, i.name, env.install_dirs[i])

        out.write_literal('\n')

        self._write_field(out, 'Name', data['desc_name'])
        self._write_field(out, 'Description', data['desc'])
        self._write_field(out, 'URL', data['url'])
        self._write_field(out, 'Version', data['version'])
        self._write_field(out, 'Requires', data['requires'], Syntax.shell,
                          delim=literal(', '))
        self._write_field(out, 'Requires.private', data['requires_private'],
                          Syntax.shell, delim=literal(', '))
        self._write_field(out, 'Conflicts', data['conflicts'],
                          Syntax.shell, delim=literal(', '))
        self._write_field(out, 'Cflags', data['cflags'], Syntax.shell)
        self._write_field(out, 'Libs', data['ldflags'], Syntax.shell)
        self._write_field(out, 'Libs.private', data['ldflags_private'],
                          Syntax.shell)

    def _process_inputs(self, env):
        desc_name = self.desc_name or self.name
        includes = self.includes or []
        libs = self.libs or []
        libs_private = self.libs_private or []
        requires, extra = self.requires or [RequirementSet(), []]
        requires_private, extra_private = (self.requires_private or
                                           [RequirementSet(), []])
        conflicts = self.conflicts or RequirementSet()

        fwd_ldflags = opts.option_list(
            getattr(i, 'forward_opts', {}).get('link_options', [])
            for i in chain(libs, libs_private)
        )

        # Add all the (unique) dependent libs to libs_private, unless they're
        # already in libs.
        fwd = chain.from_iterable(
            i.forward_opts['libs'] if hasattr(i, 'forward_opts') else []
            for i in chain(libs, libs_private)
        )
        libs_private = uniques(chain(
            (i for i in fwd if i not in libs), libs_private
        ))

        # Get the package dependencies for all the libs (public and private)
        # that were passed in.
        auto_requires, auto_extra = self._filter_packages(chain.from_iterable(
            recursive_walk(i, 'package_deps', 'install_deps')
            for i in chain(libs, libs_private)
        ))

        requires_private.update(auto_requires)
        requires.merge_from(requires_private)

        # Get the compiler and linker to use for generating flags.
        builder = env.builder(self.lang)
        compiler = builder.compiler
        linker = builder.linker('executable')

        compile_options = opts.option_list(
            (opts.include_dir(installify(i, cross=env)) for i in includes),
            self.options
        )
        link_options = opts.option_list(
            (opts.lib(installify(i.all[0], cross=env)) for i in libs),
            self.link_options
        )
        link_options_private = opts.option_list(
            (opts.lib(installify(i.all[0], cross=env)) for i in libs_private),
            fwd_ldflags, self.link_options_private
        )

        # Add the options from each of the system packages.
        for pkg in extra:
            compile_options.extend(pkg.compile_options(compiler))
            link_options.extend(pkg.link_options(linker))
        for pkg in chain(extra_private, auto_extra):
            compile_options.extend(pkg.compile_options(compiler))
            link_options_private.extend(pkg.link_options(linker))

        return {
            'desc_name': desc_name,
            'desc': self.desc or '{} library'.format(desc_name),
            'url': self.url,
            'version': self.version,

            'requires': requires.split(single=True),
            'requires_private': requires_private.split(single=True),
            'conflicts': conflicts.split(),

            'cflags': compiler.flags(compile_options, mode='pkg-config'),
            'ldflags': (linker.flags(link_options, mode='pkg-config') +
                        linker.lib_flags(link_options, mode='pkg-config')),
            'ldflags_private': (
                linker.flags(link_options_private, mode='pkg-config') +
                linker.lib_flags(link_options_private, mode='pkg-config')
            )
        }

    @staticmethod
    def _filter_packages(packages):
        pkg_config = RequirementSet()
        system = []
        for i in packages:
            if isinstance(i, string_types):
                pkg_config.add(Requirement(i))
            elif isinstance(i, (tuple, list)):
                pkg_config.add(Requirement(*i))
            elif isinstance(i, PkgConfigPackage):
                pkg_config.add(Requirement(i.name, i.specifier))
            elif isinstance(i, CommonPackage):
                system.append(i)
            else:
                raise TypeError('unsupported package type: {}'.format(type(i)))
        return pkg_config, uniques(system)


@builtin.function('builtins', 'build_inputs', 'env')
def pkg_config(builtins, build, env, name=None, **kwargs):
    if can_install(env):
        build['pkg_config'].append(
            PkgConfigInfo(builtins, build, name, **kwargs)
        )


@builtin.post('builtins', 'build_inputs', 'env')
def finalize_pkg_config(builtins, build, env):
    install = build['install']
    defaults = {
        'name': build['project'].name,
        'version': build['project'].version or '0.0',

        # Get all the explicitly-installed headers/libraries.
        'includes': [i for i in install.explicit
                     if isinstance(i, (HeaderFile, HeaderDirectory))],
        'libs': [i for i in install.explicit
                 if isinstance(i, (Library, DualUseLibrary))],
    }

    for info in build['pkg_config']:
        if not info.auto_fill:
            continue
        for key, value in iteritems(defaults):
            if getattr(info, key) is None:
                setattr(info, key, value)

    # Make sure we don't have any duplicate pkg-config packages.
    dupes = Counter(i.name for i in build['pkg_config'])
    for name, count in iteritems(dupes):
        if count > 1:
            raise ValueError("duplicate pkg-config package '{}'".format(name))

    for info in build['pkg_config']:
        with make_immediate_file(build, env, info.output) as out:
            info.write(out, env)
            builtins['install'](info.output)



from subprocess import CalledProcessError  # noqa: F401


class PackageResolutionError(Exception):
    pass


class VersionError(Exception):
    pass


class PackageVersionError(PackageResolutionError, VersionError):
    pass



import functools
from itertools import chain
from six import iteritems, string_types

from .iterutils import isiterable, iterate

__all__ = ['objectify', 'hashify', 'memoize']


def objectify(thing, valid_type, creator=None, in_type=string_types,
              **kwargs):
    if creator is None:
        creator = valid_type

    if isinstance(thing, valid_type):
        return thing
    elif not isinstance(thing, in_type):
        gen = (i.__name__ for i in chain([valid_type], iterate(in_type)))
        raise TypeError('expected {}; but got {}'.format(
            ', '.join(gen), type(thing).__name__
        ))
    else:
        return creator(thing, **kwargs)


def convert_one(kwargs, key, fn, **fn_kwargs):
    if kwargs.get(key):
        kwargs[key] = fn(kwargs[key], **fn_kwargs)
    else:
        kwargs[key] = None


def convert_each(kwargs, key, fn, **fn_kwargs):
    kwargs[key] = [fn(i, **fn_kwargs) for i in iterate(kwargs.get(key))]


def hashify(thing):
    if isinstance(thing, dict):
        return frozenset((hashify(k), hashify(v)) for k, v in iteritems(thing))
    elif isiterable(thing):
        return tuple(hashify(i) for i in thing)
    return thing


def memoize(fn):
    cache = {}

    @functools.wraps(fn)
    def wrapper(*args, **kwargs):
        key = (hashify(args), hashify(kwargs))
        if key in cache:
            return cache[key]
        result = cache[key] = fn(*args, **kwargs)
        return result

    def reset():
        cache.clear()

    wrapper._reset = reset
    return wrapper



import sys

from . import tool
from .common import SimpleCommand
from ..file_types import SourceFile
from ..languages import known_langs
from ..path import abspath

with known_langs.make('lua') as x:
    x.vars(runner='LUA')
    x.exts(source=['.lua'])

with known_langs.make('perl') as x:
    x.vars(runner='PERL')
    x.exts(source=['.pl'])

with known_langs.make('python') as x:
    x.vars(runner='PYTHON')
    x.exts(source=['.py'])

with known_langs.make('ruby') as x:
    x.vars(runner='RUBY')
    x.exts(source=['.rb'])


class ScriptCommand(SimpleCommand):
    def run_arguments(self, file):
        if isinstance(file, SourceFile):
            return self(file)
        raise TypeError('expected a source file for {} to run'
                        .format(self.rule_name))


def make_script_tool(lang, default_command):
    @tool(lang, lang=lang)
    class ScriptTool(ScriptCommand):
        def __init__(self, env):
            ScriptCommand.__init__(self, env, name=lang,
                                   env_var=known_langs[lang].var('runner'),
                                   default=default_command)

        def _call(self, cmd, file):
            return cmd + [file]

    ScriptTool.__name__ = lang.title()
    return ScriptTool


Lua = make_script_tool('lua', 'lua')
Perl = make_script_tool('perl', 'perl')
Python = make_script_tool('python', abspath(sys.executable))
Ruby = make_script_tool('ruby', 'ruby')



from itertools import chain

from . import builder
from .. import options as opts, safe_str, shell
from .common import BuildCommand, Builder, choose_builder
from ..file_types import SourceFile
from ..iterutils import iterate
from ..languages import known_langs
from ..path import Path
from ..versioning import detect_version

with known_langs.make('lex') as x:
    x.vars(compiler='LEX', flags='LFLAGS')
    x.exts(source=['.l'])

_posix_cmds = ['lex', 'flex']
_windows_cmds = ['lex', 'flex', 'win_flex']


@builder('lex')
def lex_builder(env):
    cmds = (_windows_cmds if env.host_platform.family == 'windows'
            else _posix_cmds)
    return choose_builder(env, known_langs['lex'], cmds, (LexBuilder,))


class LexBuilder(Builder):
    def __init__(self, env, langinfo, command, version_output):
        Builder.__init__(self, langinfo.name,
                         *self._parse_brand(version_output))

        name = langinfo.var('compiler').lower()
        lflags_name = langinfo.var('flags').lower()
        lflags = shell.split(env.getvar(langinfo.var('flags'), ''))

        self.transpiler = LexCompiler(self, env, name, name, command,
                                      flags=(lflags_name, lflags))

    @staticmethod
    def _parse_brand(version_output):
        if 'flex' in version_output:
            return 'flex', detect_version(version_output)
        return 'unknown', None

    @staticmethod
    def check_command(env, command):
        return env.execute(command + ['--version'], stdout=shell.Mode.pipe,
                           stderr=shell.Mode.devnull)


class LexCompiler(BuildCommand):
    @property
    def deps_flavor(self):
        return None

    @property
    def needs_libs(self):
        return False

    def _call(self, cmd, input, output, flags=None):
        return list(chain(
            cmd, iterate(flags), ['-o', output, input]
        ))

    def _output_lang(self, options):
        filtered = options.filter(opts.lang) if options else None
        if filtered:
            lang = filtered[-1].value
            if lang not in ('c', 'c++'):
                raise ValueError('only c and c++ supported')
            return lang
        return 'c'

    def default_name(self, input, context):
        options = getattr(context, 'user_options', None)
        lang = known_langs[self._output_lang(options)]
        return input.path.stripext('.yy' + lang.default_ext('source')).suffix

    def output_file(self, name, context):
        options = getattr(context, 'user_options', None)
        lang = self._output_lang(options)
        return SourceFile(Path(name), lang)

    def flags(self, options, output=None, mode='normal'):
        flags = []
        for i in options:
            if isinstance(i, opts.define):
                if i.value:
                    flags.append('-D' + i.name + '=' + i.value)
                else:
                    flags.append('-D' + i.name)
            elif isinstance(i, opts.warning):
                for j in i.value:
                    if j == opts.WarningValue.disable:
                        flags.append('-w')
                    else:
                        raise ValueError('unsupported warning level {!r}'
                                         .format(j))
            elif isinstance(i, opts.lang):
                if i.value == 'c++':
                    flags.append('--c++')
            elif isinstance(i, safe_str.stringy_types):
                flags.append(i)
            else:
                raise TypeError('unknown option type {!r}'.format(type(i)))
        return flags



from . import tool
from .common import SimpleCommand
from .. import safe_str


@tool('patchelf')
class PatchElf(SimpleCommand):
    def __init__(self, env):
        SimpleCommand.__init__(self, env, name='patchelf', env_var='PATCHELF',
                               default='patchelf')

    def _call(self, cmd, file, rpath=None):
        if rpath:
            return cmd + ['--set-rpath', safe_str.join(rpath, ':'), file]



from collections import OrderedDict
from itertools import chain
from six import iteritems, itervalues

from .path import Path, Root
from .file_types import File, Node
from .iterutils import iterate, listify, unlistify
from .objutils import objectify

_build_inputs = {}


def build_input(name):
    def wrapper(fn):
        if name in _build_inputs:  # pragma: no cover
            raise ValueError("'{}' already registered".format(name))
        _build_inputs[name] = fn
        return fn
    return wrapper


class Edge(object):
    def __init__(self, build, output, final_output=None, extra_deps=None,
                 description=None):
        self.description = description
        self.raw_output = output
        self.output = listify(output)
        for i in self.output:
            i.creator = self

        self.public_output = unlistify([
            i for i in listify(final_output) or self.output if not i.private
        ])

        def make(name):
            return build.add_source(File(Path(name, Root.srcdir)))
        self.extra_deps = [objectify(i, Node, make)
                           for i in iterate(extra_deps)]
        build.add_edge(self)


class BuildInputs(object):
    def __init__(self, env, bfgpath, extra_bootstrap=[]):
        self._sources = OrderedDict()
        self.bootstrap_paths = []
        self._edges = []
        self._extra_targets = []
        self._extra_inputs = {}

        self.bfgpath = bfgpath
        for i in chain([bfgpath], extra_bootstrap):
            self.bootstrap_paths.append(i)
            self.add_source(File(i))

        for name, fn in iteritems(_build_inputs):
            self._extra_inputs[name] = fn(self, env)

    def add_source(self, source):
        self._sources[source.path] = source
        return source

    def add_edge(self, edge):
        self._edges.append(edge)
        return edge

    def add_target(self, target):
        self._extra_targets.append(target)
        return target

    def sources(self):
        return itervalues(self._sources)

    def targets(self):
        return chain(
            chain.from_iterable(i.output for i in self._edges),
            self._extra_targets
        )

    def edges(self):
        return iter(self._edges)

    def __getitem__(self, key):
        return self._extra_inputs[key]

    def __setitem__(self, key, value):
        self._extra_inputs[key] = value



import enum
from collections import namedtuple
from six import add_metaclass, string_types
from six.moves import zip

from . import path, safe_str
from .iterutils import isiterable, iterate
from .file_types import *
from .platforms.framework import Framework


class option_list(object):
    def __init__(self, *args):
        self._options = []
        self.collect(*args)

    def append(self, option):
        if ( isinstance(option, safe_str.stringy_types) or
             not any(option.matches(i) for i in self._options) ):
            self._options.append(option)

    def extend(self, options):
        for i in options:
            self.append(i)

    def collect(self, *args):
        for i in args:
            if isiterable(i):
                for j in i:
                    self.collect(j)
            elif i is not None:
                self.append(i)

    def copy(self):
        return option_list(self._options)

    def filter(self, type):
        return option_list(i for i in self._options if isinstance(i, type))

    def __iter__(self):
        return iter(self._options)

    def __len__(self):
        return len(self._options)

    def __getitem__(self, key):
        if isinstance(key, slice):
            return type(self)(self._options[key])
        return self._options[key]

    def __setitem__(self, key, value):
        self._options[key] = value

    def __eq__(self, rhs):
        return type(self) == type(rhs) and self._options == rhs._options

    def __ne__(self, rhs):
        return not (self == rhs)

    def __nonzero__(self):
        return self.__bool__()

    def __bool__(self):
        return bool(self._options)

    def __repr__(self):
        return '<option_list({})>'.format(repr(self._options))

    def __add__(self, rhs):
        x = self.copy()
        x += rhs
        return x

    def __iadd__(self, rhs):
        if not isinstance(rhs, option_list):
            raise TypeError('expected an option_list, got a {!r}'
                            .format(type(rhs)))
        self.extend(rhs)
        return self


variadic = namedtuple('variadic', ['type'])


# XXX: This is a separate function to make Python 2.7.8 and earlier happy. For
# details, see <https://bugs.python.org/issue21591>.
def _make_init(slots, attrs):
    exec('def __init__(self, {0}):\n    self._init({0})'
         .format(', '.join(slots)), globals(), attrs)


class OptionMeta(type):
    @staticmethod
    def _make_args(slots, types):
        has_variadic = False
        for s, t in zip(slots, types):
            assert not has_variadic
            if isinstance(t, variadic):
                has_variadic = True
                yield '*' + s
            else:
                yield s

    def __new__(cls, name, bases, attrs):
        fields = attrs.pop('_fields', [])
        slots = tuple(i[0] if isiterable(i) else i for i in fields)
        types = tuple(i[1] if isiterable(i) else None for i in fields)
        attrs.update({'__slots__': [i.replace('*', '') for i in slots],
                      '_types': types})

        if '__init__' not in attrs:
            _make_init(cls._make_args(slots, types), attrs)

        return type.__new__(cls, name, bases, attrs)

    def __init__(cls, name, bases, attrs):
        is_root = not any(type(i) == OptionMeta for i in bases)
        if is_root:
            cls.registry = {}
        else:
            cls.registry[name] = cls

        type.__init__(cls, name, bases, attrs)


@add_metaclass(OptionMeta)
class Option(object):
    def _init(self, *args):
        def check_type(v, t):
            if not t or isinstance(v, t):
                return v
            elif isinstance(v, string_types) and issubclass(t, enum.Enum):
                try:
                    return t[v]
                except Exception:
                    raise ValueError('invalid {} {!r}'.format(t.__name__, v))
            else:
                raise TypeError('expected {}; but got {}'.format(
                    ', '.join(i.__name__ for i in iterate(t)), type(v).__name__
                ))

        i = 0
        for k, t in zip(self.__slots__, self._types):
            assert i < len(args)
            if isinstance(t, variadic):
                v = [check_type(x, t.type) for x in args[i:]]
                i = len(args)
            else:
                v = check_type(args[i], t)
                i += 1
            setattr(self, k, v)

    def matches(self, rhs):
        return self == rhs

    def __eq__(self, rhs):
        return type(self) == type(rhs) and all(
            getattr(self, i) == getattr(rhs, i) for i in self.__slots__
        )

    def __ne__(self, rhs):
        return not (self == rhs)

    def __repr__(self):
        return '<{}({})>'.format(self.__class__.__name__, ', '.join(
            repr(getattr(self, i)) for i in self.__slots__
        ))


class OptionEnum(enum.Enum):
    def __repr__(self):
        return repr(self.name)


def option(name, fields=[]):
    return type(name, (Option,), {'_fields': fields})


def variadic_option(name, type=None):
    return option(name, [('value', variadic(type))])


# Compilation options
include_dir = option('include_dir', [('directory', HeaderDirectory)])
std = option('std', [('value', str)])
pic = option('pic')
pch = option('pch', [('header', PrecompiledHeader)])
sanitize = option('sanitize')

WarningValue = OptionEnum('WarningValue', ['disable', 'all', 'extra', 'error'])
warning = variadic_option('warning', WarningValue)


class define(Option):
    _fields = [ ('name', str),
                ('value', (str, type(None))) ]

    def __init__(self, name, value=None):
        Option._init(self, name, value)


# Link options
lib_dir = option('lib_dir', [('directory', Directory)])
lib = option('lib', [('library', (Library, Framework, str))])
rpath_dir = option('rpath_dir', [('path', path.BasePath)])
rpath_link_dir = option('rpath_link_dir', [('path', path.BasePath)])
lib_literal = option('lib_literal', [('value', safe_str.stringy_types)])
module_def = option('module_def', [('value', ModuleDefFile)])
entry_point = option('entry_point', [('value', str)])

# General options
debug = option('debug')
lang = option('lang', [('value', str)])
pthread = option('pthread')

OptimizeValue = OptionEnum('OptimizeValue', ['disable', 'size', 'speed',
                                             'linktime'])
optimize = variadic_option('optimize', OptimizeValue)



# This is an extremely simplistic argument parser designed to understand
# arguments formatted for Visual Studio tools. It's used to parse user
# arguments so they can be correctly inserted into MSBuild files.


class ArgumentParser(object):
    _argument_info = {}

    def __init__(self, prefix_chars='/-', value_delim=':'):
        self.prefix_chars = prefix_chars
        self.value_delim = value_delim
        self._options = []
        self._short_names = {}
        self._long_names = {}
        self._unnamed_dest = None

    @classmethod
    def handler(cls, type):
        def wrapper(thing):
            cls._argument_info[type] = thing
            return thing
        return wrapper

    def add(self, *args, **kwargs):
        dest = kwargs.pop('dest', args[0][1:])
        type = kwargs.pop('type', bool)
        info = self._argument_info[type](dest, **kwargs)

        for i in args:
            if i[0] not in self.prefix_chars:
                raise ValueError('names must begin with a prefix char')

            if len(i) == 2:
                if i in self._short_names:
                    raise ValueError('{!r} already defined'.format(i))
                self._short_names[i] = info
            else:
                if i[:2] in self._short_names:
                    raise ValueError('{!r} collides with {!r}'
                                     .format(i, i[:2]))
                if i in self._long_names:
                    raise ValueError('{!r} already defined'.format(i))
                self._long_names[i] = info

        self._options.append(info)
        return info

    def add_unnamed(self, dest):
        self._unnamed_dest = dest

    def parse_known(self, args):
        result = {i.name: i.default() for i in self._options if i.name}
        if self._unnamed_dest:
            result[self._unnamed_dest] = []
        extra = []

        args = iter(args)
        while True:
            i = next(args, None)
            if i is None:
                break

            info = None
            if i[0] in self.prefix_chars:
                key, value = i[:2], i[2:]
                if key in self._short_names:
                    info = self._short_names[key]
                    if info.takes_value:
                        if not value:
                            try:
                                value = next(args)
                            except StopIteration:
                                raise ValueError('expected value for option')
                    elif value:
                        raise ValueError('no value expected for option')
                else:
                    key, colon, value = i.partition(self.value_delim)
                    if key in self._long_names:
                        info = self._long_names[key]
                        if not info.takes_value and colon:
                            raise ValueError('no value expected for option')
                        elif info.takes_value and not value:
                            raise ValueError('expected value for option')
            elif self._unnamed_dest:
                result[self._unnamed_dest].append(i)
                continue

            if info:
                info.fill_value(result, key, value)
                continue
            extra.append(i)

        return result, extra


class ArgumentInfo(object):
    def __init__(self, name):
        self.name = name

    def default(self):
        return None

    @property
    def takes_value(self):
        return True


@ArgumentParser.handler('key')
class KeyArgumentInfo(ArgumentInfo):
    def fill_value(self, results, key, value):
        results[self.name] = key


@ArgumentParser.handler('alias')
class AliasArgumentInfo(ArgumentInfo):
    def __init__(self, name, base, value=None):
        ArgumentInfo.__init__(self, None)
        self.base = base
        self.value = value
        if self.value is not None and not self.base.takes_value:
            raise TypeError('base argument does not take a value')

    def fill_value(self, results, key, value):
        if self.value is not None:
            value = self.value
        self.base.fill_value(results, key, value)

    @property
    def takes_value(self):
        return self.value is None and self.base.takes_value


@ArgumentParser.handler(bool)
class BoolArgumentInfo(ArgumentInfo):
    def __init__(self, name, value=True):
        ArgumentInfo.__init__(self, name)
        self.value = value

    def fill_value(self, results, key, value):
        results[self.name] = self.value

    @property
    def takes_value(self):
        return False


@ArgumentParser.handler(str)
class StrArgumentInfo(ArgumentInfo):
    def __init__(self, name):
        ArgumentInfo.__init__(self, name)

    def fill_value(self, results, key, value):
        results[self.name] = value


@ArgumentParser.handler(list)
class ListArgumentInfo(ArgumentInfo):
    def __init__(self, name):
        ArgumentInfo.__init__(self, name)

    def default(self):
        return []

    def fill_value(self, results, key, value):
        results[self.name].append(value)


@ArgumentParser.handler(dict)
class DictArgumentInfo(ArgumentInfo):
    def __init__(self, name):
        ArgumentInfo.__init__(self, name)
        self._short_names = {}
        self._long_names = {}
        self._options = []

    def add(self, *args, **kwargs):
        dest = kwargs.pop('dest', args[0])
        type = kwargs.pop('type', 'key')
        info = ArgumentParser._argument_info[type](dest, **kwargs)

        if type in (bool, 'key'):
            for i in args:
                self._long_names[i] = info
        else:
            for i in args:
                if len(i) != 1:
                    raise ValueError('short names should be one character')
                self._short_names[i] = info

        self._options.append(info)
        return info

    def default(self):
        return {i.name: i.default() for i in self._options}

    def fill_value(self, results, key, value):
        subkey, subvalue = value[:1], value[1:]
        if subkey in self._short_names:
            info = self._short_names[subkey]
            info.fill_value(results[self.name], subkey, subvalue)
        elif value in self._long_names:
            info = self._long_names[value]
            info.fill_value(results[self.name], value, None)
        else:
            raise ValueError('unexpected value for option')



import os.path
import warnings
from collections import defaultdict
from itertools import chain
from six.moves import filter as ifilter

from . import builtin
from .. import options as opts
from .file_types import static_file
from ..backends.make import writer as make
from ..backends.ninja import writer as ninja
from ..build_inputs import build_input, Edge
from ..file_types import *
from ..iterutils import (first, flatten, iterate, listify, merge_dicts,
                         merge_into_dict, slice_dict, uniques)
from ..languages import known_langs
from ..objutils import convert_each, convert_one
from ..shell import posix as pshell

build_input('link_options')(lambda build_inputs, env: {
    'dynamic': defaultdict(list), 'static': defaultdict(list)
})


class Link(Edge):
    msbuild_output = True
    extra_kwargs = ()

    def __init__(self, build, env, name, files, libs, packages, link_options,
                 entry_point=None, lang=None, extra_deps=None,
                 description=None):
        self.name = self.__name(name)

        self.user_libs = libs
        forward_opts = self.__get_forward_opts(self.user_libs)
        self.libs = self.user_libs + forward_opts.get('libs', [])

        self.user_packages = packages
        self.packages = self.user_packages + forward_opts.get('packages', [])

        self.user_files = files
        self.files = self.user_files + flatten(
            getattr(i, 'extra_objects', []) for i in self.user_files
        )

        if ( len(self.files) == 0 and
             not any(isinstance(i, WholeArchive) for i in self.user_libs) ):
            raise ValueError('need at least one source file')

        self.user_options = link_options
        if entry_point:
            self.entry_point = entry_point

        formats = uniques(i.format for i in chain(self.files, self.libs,
                                                  self.packages))
        if len(formats) > 1:
            raise ValueError('cannot link multiple object formats')

        self.langs = uniques(chain(
            (i.lang for i in self.files if i.lang is not None),
            (j for i in self.libs for j in iterate(i.lang))
        ))
        if not lang and not self.langs:
            raise ValueError('unable to determine language')

        search_langs = [lang] if lang else self.langs
        self.linker = self.__find_linker(env, formats[0], search_langs)

        # Forward any necessary options to the compile step.
        if hasattr(self.linker, 'compile_options'):
            compile_opts = self.linker.compile_options(self)
        else:
            compile_opts = opts.option_list()
        compile_opts.extend(forward_opts.get('compile_options', []))
        for i in self.files:
            if hasattr(i.creator, 'add_extra_options'):
                i.creator.add_extra_options(compile_opts)

        extra_options = self.linker.pre_build(build, name, self)

        output = self.linker.output_file(name, self)
        primary = first(output)

        primary.package_deps.extend(self.packages)

        self._fill_options(env, extra_options, forward_opts, output)

        options = self.options
        public_output = self.linker.post_build(build, options, output, self)
        primary.post_install = self.linker.post_install(options, output, self)

        Edge.__init__(self, build, output, public_output, extra_deps,
                      description)

        build['defaults'].add(primary)

    @classmethod
    def convert_args(cls, builtins, build, name, files, kwargs):
        lang = kwargs.get('lang')
        src_lang = known_langs[lang].src_lang if lang else None
        kwargs['lang'] = src_lang

        convert_each(kwargs, 'libs', builtins['library'],
                     kind=cls._preferred_lib, lang=src_lang)
        convert_each(kwargs, 'packages', builtins['package'], lang=src_lang)

        kwargs['link_options'] = pshell.listify(kwargs.get('link_options'),
                                                type=opts.option_list)

        intdir = ('{}.int/'.format(cls.__name(name))
                  if build['project']['intermediate_dirs'] else None)
        intdir = kwargs.pop('intermediate_dir', intdir)

        files = builtins['object_files'](
            files, includes=kwargs.pop('includes', None),
            pch=kwargs.pop('pch', None),
            options=kwargs.pop('compile_options', None),
            libs=kwargs['libs'], packages=kwargs['packages'], lang=lang,
            directory=intdir
        )

        return files, kwargs

    @classmethod
    def __name(cls, name):
        head, tail = os.path.split(name)
        return os.path.join(head, cls._prefix + tail)

    @staticmethod
    def __get_forward_opts(libs):
        result = {}

        def accumulate(libs):
            for i in libs:
                if hasattr(i, 'forward_opts'):
                    merge_into_dict(result, i.forward_opts)
                    accumulate(i.forward_opts.get('libs', []))

        accumulate(libs)
        return result

    def __find_linker(self, env, format, langs):
        for i in langs:
            linker = env.builder(i).linker(self.mode)
            if linker.can_link(format, langs):
                return linker
        raise ValueError('unable to find linker')


class DynamicLink(Link):
    desc_verb = 'link'
    base_mode = 'dynamic'
    mode = 'executable'
    msbuild_mode = 'Application'
    _preferred_lib = 'shared'
    _prefix = ''

    extra_kwargs = ('module_defs',)

    def __init__(self, *args, **kwargs):
        self.module_defs = kwargs.pop('module_defs', None)
        Link.__init__(self, *args, **kwargs)

    @classmethod
    def convert_args(cls, builtins, build, name, files, kwargs):
        convert_one(kwargs, 'module_defs', builtins['module_def_file'])
        return super(DynamicLink, cls).convert_args(builtins, build, name,
                                                    files, kwargs)

    @property
    def options(self):
        return self._internal_options + self.user_options

    @property
    def flags(self):
        return self.linker.flags(self.options, self.raw_output)

    @property
    def lib_flags(self):
        return self.linker.lib_flags(self.options)

    def _fill_options(self, env, extra_options, forward_opts, output):
        linkers = (env.builder(i).linker(self.mode) for i in self.langs)
        self._internal_options = opts.option_list(
            opts.module_def(self.module_defs) if self.module_defs else None,
            (i.link_options(self.linker) for i in self.packages)
        )

        if self.linker.needs_libs:
            self._internal_options.collect(
                (i.always_libs(i is self.linker) for i in linkers),
                (opts.lib(i) for i in self.libs)
            )

        self._internal_options.extend(extra_options)
        self._internal_options.extend(forward_opts.get('link_options', []))

        first(output).runtime_deps.extend(
            i.runtime_file for i in self.libs if i.runtime_file
        )


class SharedLink(DynamicLink):
    desc_verb = 'shared-link'
    mode = 'shared_library'
    msbuild_mode = 'DynamicLibrary'
    _prefix = 'lib'

    extra_kwargs = DynamicLink.extra_kwargs + ('version', 'soversion')

    def __init__(self, *args, **kwargs):
        self.version = kwargs.pop('version', None)
        self.soversion = kwargs.pop('soversion', None)
        if (self.version is None) != (self.soversion is None):
            raise ValueError('specify both version and soversion or neither')
        DynamicLink.__init__(self, *args, **kwargs)


class StaticLink(Link):
    desc_verb = 'static-link'
    base_mode = 'static'
    mode = 'static_library'
    msbuild_mode = 'StaticLibrary'
    _preferred_lib = 'static'
    _prefix = 'lib'

    extra_kwargs = ('static_link_options',)

    def __init__(self, *args, **kwargs):
        self.user_static_options = kwargs.pop('static_link_options', None)
        Link.__init__(self, *args, **kwargs)

    @classmethod
    def convert_args(cls, builtins, build, name, files, kwargs):
        kwargs['static_link_options'] = pshell.listify(
            kwargs.get('static_link_options'), type=opts.option_list
        )
        return super(StaticLink, cls).convert_args(builtins, build, name,
                                                   files, kwargs)

    @property
    def options(self):
        return self._internal_options + self.user_static_options

    @property
    def flags(self):
        # Only pass the static-link options to the static linker. The other
        # options are forwarded on to the dynamic linker when this library is
        # used.
        return self.linker.flags(self.options, self.raw_output)

    def _fill_options(self, env, extra_options, forward_opts, output):
        self._internal_options = extra_options

        primary = first(output)
        primary.forward_opts = {
            'compile_options': opts.option_list(),
            # Don't include libs in link_options, since later link steps will
            # handle adding them to their LDFLAGS.
            'link_options': self.user_options,
            'libs': self.libs,
            'packages': self.user_packages,
        }
        if hasattr(self.linker, 'forwarded_compile_options'):
            primary.forward_opts['compile_options'].extend(
                self.linker.forwarded_compile_options(self)
            )

        primary.linktime_deps.extend(self.user_libs)


@builtin.function('builtins', 'build_inputs', 'env')
@builtin.type(Executable)
def executable(builtins, build, env, name, files=None, **kwargs):
    if files is None and 'libs' not in kwargs:
        dist = kwargs.pop('dist', True)
        params = [('format', env.target_platform.object_format),
                  ('lang', build['project']['lang'])]
        return static_file(build, Executable, name, dist, params, kwargs)
    files, kwargs = DynamicLink.convert_args(builtins, build, name, files,
                                             kwargs)
    return DynamicLink(build, env, name, files, **kwargs).public_output


@builtin.function('builtins', 'build_inputs', 'env')
@builtin.type(SharedLibrary, extra_in_type=DualUseLibrary)
def shared_library(builtins, build, env, name, files=None, **kwargs):
    if isinstance(name, DualUseLibrary):
        if files is not None or not set(kwargs.keys()) <= {'format', 'lang'}:
            raise TypeError('unexpected arguments')
        return name.shared

    if files is None and 'libs' not in kwargs:
        # XXX: What to do for pre-built shared libraries for Windows, which has
        # a separate DLL file?
        dist = kwargs.pop('dist', True)
        params = [('format', env.target_platform.object_format),
                  ('lang', build['project']['lang'])]
        return static_file(build, SharedLibrary, name, dist, params, kwargs)
    files, kwargs = SharedLink.convert_args(builtins, build, name, files,
                                            kwargs)
    return SharedLink(build, env, name, files, **kwargs).public_output


@builtin.function('builtins', 'build_inputs', 'env')
@builtin.type(StaticLibrary, extra_in_type=DualUseLibrary)
def static_library(builtins, build, env, name, files=None, **kwargs):
    if isinstance(name, DualUseLibrary):
        if files is not None or not set(kwargs.keys()) <= {'format', 'lang'}:
            raise TypeError('unexpected arguments')
        return name.static

    if files is None and 'libs' not in kwargs:
        dist = kwargs.pop('dist', True)
        params = [('format', env.target_platform.object_format),
                  ('lang', build['project']['lang'])]
        return static_file(build, StaticLibrary, name, dist, params, kwargs)
    files, kwargs = StaticLink.convert_args(builtins, build, name, files,
                                            kwargs)
    return StaticLink(build, env, name, files, **kwargs).public_output


@builtin.function('builtins', 'build_inputs', 'env')
@builtin.type(Library, extra_in_type=DualUseLibrary)
def library(builtins, build, env, name, files=None, **kwargs):
    explicit_kind = False

    if 'kind' in kwargs:
        explicit_kind = True
        kind = kwargs.pop('kind')
    elif env.library_mode.shared and env.library_mode.static:
        kind = 'dual'
    elif env.library_mode.shared:
        kind = 'shared'
    elif env.library_mode.static:
        kind = 'static'
    else:
        raise ValueError('unable to create library: both shared and static ' +
                         'modes disabled')

    if isinstance(name, DualUseLibrary):
        if files is not None or not set(kwargs.keys()) <= {'format', 'lang'}:
            raise TypeError('unexpected arguments')
        return name if kind == 'dual' else getattr(name, kind)

    if files is None and 'libs' not in kwargs:
        dist = kwargs.pop('dist', True)
        params = [('format', env.target_platform.object_format),
                  ('lang', build['project']['lang'])]
        file_type = StaticLibrary

        if explicit_kind:
            if kind == 'shared':
                file_type = SharedLibrary
            elif kind == 'dual':
                raise ValueError("can't create dual-use libraries from an " +
                                 "existing file")

        # XXX: Try to detect if a string refers to a shared lib?
        return static_file(build, file_type, name, dist, params, kwargs)

    shared_kwargs = slice_dict(kwargs, SharedLink.extra_kwargs)
    static_kwargs = slice_dict(kwargs, StaticLink.extra_kwargs)
    shared_kwargs.update(kwargs)
    static_kwargs.update(kwargs)

    if kind == 'dual':
        shared_files, shared_kwargs = SharedLink.convert_args(
            builtins, build, name, files, shared_kwargs
        )
        shared = SharedLink(build, env, name, shared_files, **shared_kwargs)
        if not shared.linker.builder.can_dual_link:
            warnings.warn("dual linking not supported with {}"
                          .format(shared.linker.brand))
            return shared.public_output

        static_files, static_kwargs = StaticLink.convert_args(
            builtins, build, name, shared_files, static_kwargs
        )
        static = StaticLink(build, env, name, static_files, **static_kwargs)
        return DualUseLibrary(shared.public_output, static.public_output)
    elif kind == 'shared':
        files, kw = SharedLink.convert_args(builtins, build, name, files,
                                            shared_kwargs)
        return SharedLink(build, env, name, files, **kw).public_output
    else:  # kind == 'static'
        files, kw = StaticLink.convert_args(builtins, build, name, files,
                                            static_kwargs)
        return StaticLink(build, env, name, files, **kw).public_output


@builtin.function('builtins')
@builtin.type(WholeArchive, extra_in_type=StaticLibrary)
def whole_archive(builtins, name, *args, **kwargs):
    if isinstance(name, StaticLibrary):
        if len(args) or len(kwargs):
            raise TypeError('unexpected arguments')
        return WholeArchive(name)
    else:
        return WholeArchive(builtins['static_library'](name, *args, **kwargs))


@builtin.function('build_inputs')
def global_link_options(build, options, family='native', mode='dynamic'):
    for i in iterate(family):
        build['link_options'][mode][i].extend(pshell.listify(options))


def _get_flags(backend, rule, build_inputs, buildfile):
    variables = {}
    cmd_kwargs = {}

    linker = rule.linker
    if hasattr(linker, 'flags_var'):
        global_ldflags, ldflags = backend.flags_vars(
            linker.flags_var,
            (linker.global_flags +
             linker.flags(build_inputs['link_options'][rule.base_mode]
                          [linker.family])),
            buildfile
        )
        cmd_kwargs['flags'] = ldflags
        flags = rule.flags
        if flags:
            variables[ldflags] = [global_ldflags] + flags

    if hasattr(rule.linker, 'libs_var'):
        global_ldlibs, ldlibs = backend.flags_vars(
            rule.linker.libs_var, rule.linker.global_libs, buildfile
        )
        cmd_kwargs['libs'] = ldlibs
        lib_flags = rule.lib_flags
        if lib_flags:
            variables[ldlibs] = [global_ldlibs] + lib_flags

    if hasattr(rule, 'manifest'):
        var = backend.var('manifest')
        cmd_kwargs['manifest'] = var
        variables[var] = rule.manifest

    return variables, cmd_kwargs


@make.rule_handler(StaticLink, DynamicLink, SharedLink)
def make_link(rule, build_inputs, buildfile, env):
    linker = rule.linker
    variables, cmd_kwargs = _get_flags(make, rule, build_inputs, buildfile)

    output_params = []
    if linker.num_outputs == 'all':
        output_vars = make.qvar('@')
    else:
        output_vars = []
        for i in range(linker.num_outputs):
            v = make.var(str(i + 2))
            output_vars.append(v)
            output_params.append(rule.output[i])

    recipename = make.var('RULE_{}'.format(linker.rule_name.upper()))
    if not buildfile.has_variable(recipename):
        buildfile.define(recipename, [linker(
            make.var('1'), output_vars, **cmd_kwargs
        )])

    files = rule.files
    if hasattr(linker, 'transform_input'):
        files = linker.transform_input(files)

    manifest = listify(getattr(rule, 'manifest', None))
    module_defs = listify(getattr(rule, 'module_defs', None))
    make.multitarget_rule(
        buildfile,
        targets=rule.output,
        deps=rule.files + rule.libs + module_defs + manifest + rule.extra_deps,
        order_only=make.directory_deps(rule.output),
        recipe=make.Call(recipename, files, *output_params),
        variables=variables
    )


@ninja.rule_handler(StaticLink, DynamicLink, SharedLink)
def ninja_link(rule, build_inputs, buildfile, env):
    linker = rule.linker
    variables, cmd_kwargs = _get_flags(ninja, rule, build_inputs, buildfile)
    if rule.description:
        variables['description'] = rule.description

    if linker.num_outputs == 'all':
        output_vars = ninja.var('out')
    elif linker.num_outputs == 1:
        output_vars = ninja.var('output')
        variables[output_vars] = rule.output[0]
    else:
        output_vars = []
        for i in range(linker.num_outputs):
            v = ninja.var('output{}'.format(i + 1))
            output_vars.append(v)
            variables[v] = rule.output[i]

    if hasattr(linker, 'transform_input'):
        input_var = ninja.var('input')
        variables[input_var] = linker.transform_input(rule.files)
    else:
        input_var = ninja.var('in')

    if not buildfile.has_rule(linker.rule_name):
        buildfile.rule(name=linker.rule_name, command=linker(
            input_var, output_vars, **cmd_kwargs
        ), description=rule.desc_verb + ' => ' + first(output_vars))

    manifest = listify(getattr(rule, 'manifest', None))
    module_defs = listify(getattr(rule, 'module_defs', None))
    buildfile.build(
        output=rule.output,
        rule=linker.rule_name,
        inputs=rule.files,
        implicit=rule.libs + module_defs + manifest + rule.extra_deps,
        variables=variables
    )


try:
    from .compile import CompileHeader
    from ..backends.msbuild import writer as msbuild

    def _parse_compiler_cflags(compilers, global_options):
        per_compiler_cflags = {}
        for c in compilers:
            key = c.command_var
            if key not in per_compiler_cflags:
                per_compiler_cflags[key] = c.parse_flags(msbuild.textify_each(
                    c.global_flags + c.flags(global_options[c.lang])
                ))
        return per_compiler_cflags

    def _parse_file_cflags(file, per_compiler_cflags):
        cflags = file.creator.compiler.parse_flags(
            msbuild.textify_each(file.creator.flags)
        )
        if not per_compiler_cflags:
            return cflags
        key = file.creator.compiler.command_var
        return merge_dicts(per_compiler_cflags[key], cflags)

    @msbuild.rule_handler(DynamicLink, SharedLink, StaticLink)
    def msbuild_link(rule, build_inputs, solution, env):
        if ( any(i not in ['c', 'c++'] for i in rule.langs) or
             rule.linker.flavor != 'msvc' ):
            raise ValueError('msbuild backend currently only supports c/c++ ' +
                             'with msvc')

        output = rule.output[0]

        # Parse compilation flags; if there's only one set of them (i.e. the
        # command_var is the same for every compiler), we can apply these to
        # all the files at once. Otherwise, we need to apply them to each file
        # individually so they all get the correct options.
        obj_creators = [i.creator for i in rule.files]
        compilers = uniques(i.compiler for i in obj_creators)

        per_compiler_options = _parse_compiler_cflags(
            compilers, build_inputs['compile_options']
        )
        if len(per_compiler_options) == 1:
            common_compile_options = per_compiler_options.popitem()[1]
        else:
            common_compile_options = None

        # Parse linking flags.
        ldflags = [
            rule.linker.global_flags +
            rule.linker.flags(build_inputs['link_options'][rule.base_mode]
                              [rule.linker.family]) +
            rule.flags
        ]
        if hasattr(rule.linker, 'libs_var'):
            ldflags.append(rule.linker.global_libs + rule.lib_flags)
        link_options = rule.linker.parse_flags(
            *[msbuild.textify_each(i) for i in ldflags]
        )
        if hasattr(output, 'import_lib'):
            link_options['import_lib'] = output.import_lib

        deps = chain(
            (i.creator.file for i in rule.files),
            chain.from_iterable(i.creator.include_deps for i in rule.files),
            chain.from_iterable(i.creator.extra_deps for i in rule.files),
            ifilter(None, (getattr(i.creator, 'pch_source', None)
                           for i in rule.files)),
            rule.libs, rule.extra_deps
        )

        def get_source(file):
            # Get the source file for this compilation rule; it's either a
            # regular source file or a PCH source file.
            if isinstance(file.creator, CompileHeader):
                return file.creator.pch_source
            return file.creator.file

        # Create the project file.
        project = msbuild.VcxProject(
            env, name=rule.name,
            mode=rule.msbuild_mode,
            output_file=output,
            files=[{
                'name': get_source(i),
                'options': _parse_file_cflags(i, per_compiler_options),
            } for i in rule.files],
            compile_options=common_compile_options,
            link_options=link_options,
            dependencies=solution.dependencies(deps),
        )
        solution[output] = project
except ImportError:  # pragma: no cover
    pass
